# -*- coding: future_fstrings -*-
# This module contains a set of functions and classes that are used to write text files to Disk


from pyFAT_astro.Support.modify_template import set_model_parameters, set_overall_parameters,\
                                set_fitting_parameters,get_warp_slope, update_disk_angles
from pyFAT_astro.Support.fits_functions import extract_pv
from pyFAT_astro.Support.read_functions import load_basicinfo
from pyFAT_astro.Support.fat_errors import ProgramError
import pyFAT_astro.Support.support_functions as sf
import pyFAT_astro


import copy
import numpy as np
import psutil as psu
import time
import warnings
from datetime import datetime
import traceback
import os

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import matplotlib
    matplotlib.use('pdf')
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from matplotlib.patches import Ellipse
    import matplotlib.axes as maxes
    import matplotlib.font_manager as mpl_fm
from astropy.io import fits
from astropy.wcs import WCS
# create or append to the basic ifo file
def basicinfo(Configuration,initialize = False,stage='TiRiFiC' ,
              RA=[float('NaN'),float('NaN')], DEC=[float('NaN'),float('NaN')],
              VSYS =[float('NaN'),float('NaN')], PA=[float('NaN'),float('NaN')],
              Inclination = [float('NaN'),float('NaN')], Max_Vrot = [float('NaN'),float('NaN')],
              Tot_Flux = [float('NaN'),float('NaN')], V_mask = [float('NaN'),float('NaN')],
              Distance = float('NaN') , DHI = float('NaN'), template = ['EMPTY']):

    try:
        if template[0] == 'EMPTY':
            template = {}
    except KeyError:
        pass
    if initialize:
        with open(f"{Configuration['FITTING_DIR']}{Configuration['BASE_NAME']}-Basic_Info.txt",'w') as file:
            file.write(f'''#This file contains the basic parameters of the Galaxy
#The total flux is determined by the source finder in the initial estimates and the total emission in the masked cube of the final model in the fits
#D_HI is determined as the diameter where the major axis with given PA cuts the 10^20 column of the moment0 map
# {'RA':>25s} {'DEC':>25s} {'VSYS':>20s} {'PA':>20s} {'Inclination':>20s} {'Max VRot':>20s} {'V_mask':>20s} {'Tot FLux':>20s} {'D_HI':>20s} {'Distance':>20s} {'HI Mass':>20s} {'D_HI_kpc':>20s}
# {'hh:mm:ss':>25s} {'dd:mm:ss':>25s} {'km/s':>20s} {'deg':>20s} {'deg':>20s} {'km/s':>20s} {'km/s':>20s} {'Jy':>20s} {'arcsec':>20s} {'Mpc':>20s} {'M_solar':>20s} {'kpc':>20s}
# The initial input
''')
    else:
        Vars_to_Set =  ['XPOS','YPOS','VSYS','VROT','INCL','PA','SDIS','SBR','SBR_2','Z0']
        FAT_Model = sf.load_tirific(Configuration,template,\
            Variables= Vars_to_Set,array=True )
        RA=[FAT_Model[Vars_to_Set.index('XPOS'),0],abs(Configuration['BEAM'][0]/(3600.*2.))]
        DEC=[FAT_Model[Vars_to_Set.index('YPOS'),0],abs(Configuration['BEAM'][0]/(3600.*2.))]
        VSYS =np.array([FAT_Model[Vars_to_Set.index('VSYS'),0],Configuration['CHANNEL_WIDTH']],dtype=float)
        PA=[FAT_Model[Vars_to_Set.index('PA'),0], 3.]
        Inclination = [FAT_Model[Vars_to_Set.index('INCL'),0], 3.]
        Max_Vrot = [np.max(FAT_Model[Vars_to_Set.index('VROT'),:]),np.max(FAT_Model[Vars_to_Set.index('VROT'),:])-np.min(FAT_Model[Vars_to_Set.index('VROT'),1:]) ]
        Distance = Configuration['DISTANCE']

    with open(f"{Configuration['FITTING_DIR']}{Configuration['BASE_NAME']}-Basic_Info.txt",'a') as file:
        if not initialize:
            file.write(f'''#These are the values from {stage}. \n''')
        RAhr,DEChr = sf.convertRADEC(Configuration,RA[0],DEC[0] )
        RA_c = f'{RAhr}+/-{RA[1]*3600.:0.2f}'
        DEC_c = f'{DEChr}+/-{DEC[1]*3600.:0.2f}'
        VSYS_c = f'{VSYS[0]:.2f}+/-{VSYS[1]:.2f}'
        PA_c = f'{PA[0]:.2f}+/-{PA[1]:.2f}'
        INCL_c = f'{Inclination[0]:.2f}+/-{Inclination[1]:.2f}'
        MVROT_c = f'{Max_Vrot[0]:.2f}+/-{Max_Vrot[1]:.2f}'
        Vmask_c = f'{V_mask[0]/1000.:.2f}+/-{V_mask[1]/1000.:.2f}'
        DHI_a = f'{DHI[0]:.2f}+/-{DHI[1]:.2f}'
        Dist = f'{Distance:.2f}'
        HIMass  = f'{Tot_Flux[0]*2.36E5*Distance**2:.2e}'
        DHI_k = f'{sf.convertskyangle(Configuration,DHI[0]):.2f}'
        Flux_c = f'{Tot_Flux[0]:.2f}+/-{Tot_Flux[1]:.2f}'
        file.write(f'''  {RA_c:>25s} {DEC_c:>25s} {VSYS_c:>20s} {PA_c:>20s} {INCL_c:>20s} {MVROT_c:>20s} {Vmask_c:>20s} {Flux_c:>20s} {DHI_a:>20s} {Dist:>20s} {HIMass:>20s} {DHI_k:>20s}
''')

basicinfo.__doc__ =f'''
 NAME:
    basicinfo

 PURPOSE:
    Create or update the file with the basic information of the fit

 CATEGORY:
    write_functions

 INPUTS:
    Configuration = Standard FAT configuration

 OPTIONAL INPUTS:


    initialize = False
    If true a new file with header but and first guess values will be created.
    These parameters need to be provided through the keywords below.

    stage = 'Tirific'

    RA=[float('NaN'),float('NaN')], DEC=[float('NaN'),float('NaN')]
    RA and DEC in degrees

    VSYS =[float('NaN'),float('NaN')]
    Systemic velocity in m/s

    PA=[float('NaN'),float('NaN')], Inclination = [float('NaN'),float('NaN')]
    Position angle and inclination in degrees

    Max_Vrot = [float('NaN'),float('NaN')]
    The maximum rotational velocity

    Tot_Flux = [float('NaN'),float('NaN')]
    The total flux in the source in Jy

    V_mask = [float('NaN'),float('NaN')],
    extend of the mask in the cube

    Distance = float('NaN')
    Distance to the source in Mpc

    DHI = float('NaN')
    Diameter of the 1 M_solar/pc^2 iso density surface ellipse. in arcsec

    Standard FAT dictionary with filenames

    template = ['EMPTY']
    Tirific template to obtain values from

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

# Function to write the first def file for a galaxy
def initialize_def_file(Configuration, Fits_Files,Tirific_Template,Initial_Parameters = ['EMPTY'], fit_type = 'Undefined' ):
    try:
        if Initial_Parameters[0] == 'EMPTY':
            Initial_Parameters = {}
    except KeyError:
        pass

    #First we set some basic parameters that will hardly change
    if fit_type == 'Centre_Convergence':

        #if 'VSYS' in Initial_Parameters:
        #    Initial_Parameters['VSYS'] = [x/1000. for x in Initial_Parameters['VSYS']]
        set_overall_parameters(Configuration, Fits_Files,Tirific_Template,\
            fit_type=fit_type, flux = Initial_Parameters['FLUX'][0] )
        # Then set the values for the various parameters of the model

        set_model_parameters(Configuration, Tirific_Template,Initial_Parameters )

        sf.set_limit_modifier(Configuration,Tirific_Template  )

        set_fitting_parameters(Configuration, Tirific_Template,stage = 'initial',
                                initial_estimates = Initial_Parameters )
        #if 'VSYS' in Initial_Parameters:
        #    Initial_Parameters['VSYS'] = [x*1000. for x in Initial_Parameters['VSYS']]
    elif fit_type in ['Extent_Convergence','Fit_Tirific_OSC']:
        #if 'VSYS' in Initial_Parameters and fit_type == 'Fit_Tirific_OSC':
        #    Initial_Parameters['VSYS'] = [x/1000. for x in Initial_Parameters['VSYS']]
        set_overall_parameters(Configuration, Fits_Files,Tirific_Template ,\
            fit_type=fit_type)
        Vars_to_Set =  ['XPOS','YPOS','VSYS','VROT','INCL','PA','SDIS','SBR','SBR_2','Z0']
        if fit_type == 'Fit_Tirific_OSC':
            set_model_parameters(Configuration, Tirific_Template,Initial_Parameters,stage='initialize_def_file' )

        # Finally we set how these parameters are fitted.
        sf.set_limit_modifier(Configuration,Tirific_Template )
        #get_inner_fix(Configuration,Tirific_Template )
        get_warp_slope(Configuration,Tirific_Template )
        #if Configuration['OUTER_RINGS_DOUBLED']:
        Initial_Parameters['XPOS'][1]= sf.set_limits(Initial_Parameters['XPOS'][1],Configuration['BEAM'][0]/3600.,Configuration['BEAM'][0]/3600.*5)
        Initial_Parameters['YPOS'][1]= sf.set_limits(Initial_Parameters['YPOS'][1],Configuration['BEAM'][0]/3600.,Configuration['BEAM'][0]/3600.*5)
        Initial_Parameters['VSYS'][1]= sf.set_limits(Initial_Parameters['VSYS'][1],Configuration['CHANNEL_WIDTH']/2.,Configuration['CHANNEL_WIDTH']*5)
        if Initial_Parameters['INCL'][0] >35:
            Initial_Parameters['PA'][1]= sf.set_limits(Initial_Parameters['PA'][1],1.,15)
        Initial_Parameters['INCL'][1]= sf.set_limits(Initial_Parameters['INCL'][1],3.,15)


        set_fitting_parameters(Configuration, Tirific_Template,stage = 'initialize_os',
                               initial_estimates=Initial_Parameters )

    tirific(Configuration,Tirific_Template,name = f'{fit_type}_In.def' )

initialize_def_file.__doc__ =f'''
 NAME:
    initialize_def_file

 PURPOSE:
    setup the first def file to be used in the fitting. As it is the first instance it has some special requirements.

 CATEGORY:
    write_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames
    Tirific_Template = Standard FAT Tirific Template


 OPTIONAL INPUTS:


    Initial_Parameters = ['EMPTY']
    The initial parameters to be used to set up the def file.

    fit_type = 'Undefined'
    type of fitting being done

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:"/usr/share/fonts/truetype/msttcorefonts/Times_New_Roman.ttf"
    Unspecified

 NOTE:
'''


class full_system_tracking:
    def __init__(self,Configuration):
        self.stop = False
        self.pid = os.getpid()
        self.main_pyFAT = psu.Process(self.pid)
        self.user = self.main_pyFAT.username()
        self.python = self.main_pyFAT.name()
        self.tirific = Configuration['TIRIFIC']
        self.font_file = Configuration['FONT_FILE']
        self.sofia = Configuration['SOFIA2']
        self.file = f"{Configuration['MAIN_DIRECTORY']}FAT_Resources_Used.txt"
        self.plot_name= f"{Configuration['MAIN_DIRECTORY']}pyFAT_Resources_Monitor.pdf"
        self.cpus= psu.cpu_count()
        with open(self.file,'w') as resources:
            resources.write("# This file contains an estimate of all resources used for a pyFAT run. \n")
            resources.write(f"# {'Time':20s} {'Sys CPU':>10s} {'Sys RAM':>10s} {'FAT CPU':>10s} {'FAT RAM':>10s} \n")
            resources.write(f"# {'YYYY-MM-DD hh:mm:ss':20s} {'%':>10s} {'Gb':>10s} {'%':>10s} {'Gb':>10s} \n")
        self.interval = 60 # amount of second when to do new monitor

    def start_monitoring(self):
        while not self.stop:
            try:
                self.sys_cpu= psu.cpu_percent(interval=1)
                self.sys_ram= psu.virtual_memory().used/2**30.
                self.CPU = 0.
                self.RAM = 0.
                for proc in psu.process_iter():
                    if proc.username() == self.user \
                        and proc.status() == 'running'\
                        and (proc.name() == self.python or\
                         proc.name() == self.tirific or\
                         proc.name() == self.sofia or\
                         proc.name() == 'python3'):
                        try:
                            self.CPU += proc.cpu_percent(interval=0.5)/self.cpus
                            self.RAM += (proc.memory_info()[0])/2**30.
                        except:
                            pass
                #file.write(f"{datetime.now()} CPU = {CPU} % Mem = {mem} Gb for TiRiFiC \n")
                with open(self.file,'a') as resources:
                    resources.write(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S'):20s} {self.sys_cpu:>10.1f} {self.sys_ram:>10.2f} {self.CPU:>10.1f} {self.RAM:>10.2f} \n")
            except Exception as e:
                #We do not care if something goes wrong once. We don't want the monitor to crash
                #but we would like to know what went wrong
                traceback.print_exception(type(e),e,e.__traceback__)
                pass
            time.sleep(self.interval)

    def stop_monitoring(self):
        self.stop = True
        loads = {'Time':[]}
        keys=['SCPU','SRAM','FCPU','FRAM']
        for key in keys:
            loads[key]  = []
        with open(self.file) as file:
            lines = file.readlines()
        startdate = 0
        #load data from file into dictionary
        for line in lines:
            line = line.split()
            if line[0] == '#':
                continue
            else:
                date = extract_date(f"{line[0]} {line[1]}")
            if startdate == 0:
                startdate = date
            diff = date - startdate
            time = diff.total_seconds()/(3600.)
            loads['Time'].append(time)
            for i,key in enumerate(keys):
                loads[key].append(float(line[int(2+i)]))
        #Plot the parameters
        try:
            mpl_fm.fontManager.addfont(self.font_file)
            font_name = mpl_fm.FontProperties(fname=self.font_file).get_name()
        except FileNotFoundError:
            font_name = 'DejaVu Sans'
        labelfont = {'family': font_name,
                 'weight': 'normal',
                 'size': 4}
        fig, ax1 = plt.subplots(figsize = (8,6))
        fig.subplots_adjust(left = 0.1, right = 0.9, bottom = 0.3, top = 0.7)
        ax1.plot(loads['Time'],loads['SRAM'],'b--',lw=0.5,alpha=0.5, label='System RAM')
        ax1.plot(loads['Time'],loads['FRAM'],'b-',lw=0.5,alpha=1.0, label='pyFAT RAM')
        ax1.set_ylim(0,np.max(np.array(loads['SRAM']+loads['FRAM'],dtype=float))*1.1)
        ax1.set_ylabel('RAM (Gb) ', color='b')
        ax1.tick_params(axis='y', labelcolor='b')
        ax1.set_xlabel('Run Duration (h)', color='k',zorder=5)
        ax2 = ax1.twinx()
        ax2.plot(loads['Time'],loads['SCPU'],'r--',lw=0.5,alpha=0.5, label='System CPU')
        ax2.plot(loads['Time'],loads['FCPU'],'r-',lw=0.5,alpha=1.0, label='pyFAT CPU')
        ax2.set_ylim(0,np.max(np.array(loads['SCPU']+loads['FCPU'],dtype=float))*1.1)
        ax2.set_ylabel('CPUs (%)',color='r')
        ax2.tick_params(axis='y', labelcolor='r')
        fig.savefig(self.plot_name)
        plt.close()

def make_overview_plot(Configuration,Fits_Files ):
    fit_type = Configuration['USED_FITTING']
    sf.print_log(f'''MAKE_OVERVIEW_PLOT: We are starting the overview plot.
''',Configuration,case=['debug_start'])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        cube_mod = fits.open(f"{Configuration['FITTING_DIR']}/Finalmodel/Finalmodel.fits")
        moment0_mod = fits.open(f"{Configuration['FITTING_DIR']}/Finalmodel/Finalmodel_mom0.fits")
        moment1_mod = fits.open(f"{Configuration['FITTING_DIR']}/Finalmodel/Finalmodel_mom1.fits")
        moment2_mod = fits.open(f"{Configuration['FITTING_DIR']}/Finalmodel/Finalmodel_mom2.fits")
        cube = fits.open(f"{Configuration['FITTING_DIR']}{Fits_Files['FITTING_CUBE']}")
        moment0 = fits.open(f"{Configuration['FITTING_DIR']}{Fits_Files['MOMENT0']}")
        moment1 = fits.open(f"{Configuration['FITTING_DIR']}{Fits_Files['MOMENT1']}")
        moment2 = fits.open(f"{Configuration['FITTING_DIR']}{Fits_Files['MOMENT2']}")
        channels_map = fits.open(f"{Configuration['FITTING_DIR']}{Fits_Files['CHANNEL_MAP']}")
        im_wcs = WCS(moment0[0].header).celestial

    # Open the model info
    sf.print_log(f'''MAKE_OVERVIEW_PLOT: Reading the variables from the final model
''',Configuration,case=['debug_add'])
    Vars_to_plot_short= ['RADI','XPOS','YPOS','VSYS','VROT','INCL','PA','SDIS',\
                    'SBR','Z0']
    Vars_to_plot=copy.deepcopy(Vars_to_plot_short)
    for x in Vars_to_plot_short:
        if x != 'RADI':
            Vars_to_plot.append(f'# {x}_ERR')
        if x not in ['XPOS','YPOS','VSYS']:
            Vars_to_plot.append(f'{x}_2')
            Vars_to_plot.append(f'# {x}_2_ERR')
    FAT_Model = sf.load_tirific(Configuration,\
        f"{Configuration['FITTING_DIR']}Finalmodel/Finalmodel.def",\
        Variables= Vars_to_plot,array=True )
    Extra_Model_File = f"{Configuration['FITTING_DIR']}{fit_type}/{fit_type}_Iteration_{Configuration['ITERATIONS']}.def"

    if os.path.exists(Extra_Model_File):
        Extra_Model = sf.load_tirific(Configuration,Extra_Model_File,\
            Variables= Vars_to_plot,array=True )
    else:
        Extra_Model = []
    sf.print_log(f'''MAKE_OVERVIEW_PLOT: We find the following model values.
{'':8s}{[f"{x} = {FAT_Model[i,:]}" for i,x in enumerate(Vars_to_plot)]}
''',Configuration,case=['debug_add'])

    if os.path.exists(f"{Configuration['FITTING_DIR']}ModelInput.def"):
        Input_Model = sf.load_tirific(Configuration,\
            f"{Configuration['FITTING_DIR']}ModelInput.def",\
            Variables= Vars_to_plot,array=True )
    else:
        Input_Model = []
    sof_basic_ra,sof_basic_dec, sof_basic_vsys,sof_basic_maxrot,sof_basic_pa,sof_basic_inclination,sof_basic_extent = load_basicinfo(Configuration,
        f"{Configuration['FITTING_DIR']}{Configuration['BASE_NAME']}-Basic_Info.txt",Variables=['RA','DEC','VSYS','Max VRot','PA','Inclination','D_HI'])


    #Let's start plotting
    ysize = 23.2
    xsize = 0.7*ysize
    Overview = plt.figure(2, figsize=(xsize, ysize), dpi=300, facecolor='w', edgecolor='k')

    size_factor= ysize/11.6
    size_ratio = ysize/xsize

    #stupid pythonic layout for grid spec, which means it is yx instead of xy like for normal human beings
    gs = Overview.add_gridspec(int(20*size_ratio),20)
    try:
        mpl_fm.fontManager.addfont(Configuration['FONT_FILE'])
        font_name = mpl_fm.FontProperties(fname=Configuration['FONT_FILE']).get_name()
    except FileNotFoundError:
        font_name = 'DejaVu Sans'
    labelfont = {'family': font_name,
             'weight': 'normal',
             'size': 8*size_factor}
    plt.rc('font', **labelfont)
    plt.rcParams['xtick.direction'] = 'in'
    plt.rcParams['ytick.direction'] = 'in'
# First some textual information

    if Configuration['SUB_DIR'] == './':
        name = Configuration['BASE_NAME']
    else:
        name = Configuration['SUB_DIR']
    #plt.title(f"Overview for {name}")
    ax_text = Overview.add_subplot(gs[0:1,0:21],frameon = False)
    plt.tick_params(
        axis='both',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        direction = 'in',
        bottom=False,      # ticks along the bottom edge are off
        top=False,         # ticks along the top edge are off
        labelbottom=False,
        right = False,
        left= False,
        labelleft = False)
    ax_text.text(0.5,1.0,f'''Overview for {name}''',rotation=0, va='top',ha='center', color='black',transform = ax_text.transAxes,
      bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7,fontsize=14*size_factor)
    ax_text.text(0.5,0.25,f'''The ring size used in the model is {Configuration['RING_SIZE']:.2f} x BMAJ, with BMAJ = {Configuration['BEAM'][0]:.1f} arcsec. We assumed a distance  of {Configuration['DISTANCE']:.1f} Mpc.'''
      ,rotation=0, va='top',ha='center', color='black',transform = ax_text.transAxes,
      bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7,fontsize=10*size_factor)


#-----------------------------------------------------------------Moment 0 ------------------------------------------------------
    ax_moment0 = Overview.add_subplot(gs[2:8,0:6], projection=im_wcs)
    ax_moment0.set_label('Intensity Map')
    #Comp_ax1.set_facecolor('black')
    # we need contour levels and
    min_color = 0.
    max_color = np.nanmax(moment0[0].data)*0.8
    moment0_plot = ax_moment0.imshow(moment0[0].data, origin='lower', alpha=1, vmin = min_color, vmax = max_color,cmap='hot_r' )
    moment0_plot.set_label('Intensity Map')
    plt.ylabel('DEC (J2000)')
    #Stupid python suddenly finds its own labels
    plt.xlabel('RA J2000')

    median_noise_in_map = np.sqrt(np.nanmedian(channels_map[0].data[channels_map[0].data > 0.]))*Configuration['NOISE']*Configuration['CHANNEL_WIDTH']
    mindism0 = median_noise_in_map
    mindism0 = median_noise_in_map
    #"We find this {} as the minimum of the moment0 map".format(mindism0))
    if mindism0 < 0.:
        mindism0  =abs(mindism0)*2.
    if mindism0 == 0:
        mindism0 = np.max(moment0[0].data)/64.
    #print("We find this {} as the minimum of the moment0 map".format(mindism0))
    maxdism0 = np.max(moment0[0].data) * 0.8
    if mindism0 > maxdism0:
        mindism0 = 0.1*maxdism0
    if maxdism0 < 16*mindism0:
        momlevel = np.array([1,4,8,12],dtype=float)* mindism0
    elif maxdism0 < 32*mindism0:
        momlevel = np.array([1,4,8,16,24,32],dtype=float)* mindism0
    elif maxdism0 < 256*mindism0:
        momlevel = np.array([1,4,8,32,64,128],dtype=float) * mindism0
    else:
        momlevel = np.array([0,1,2,3,4,5,6,7],dtype=float) * (maxdism0-mindism0)/7+ mindism0

    #print("We find this {} as the minimum of the moment0 map".format(mindism0))
    momlevel = np.array([x for x in momlevel if x < np.max(moment0[0].data)*0.95],dtype=float)
    if momlevel.size == 0:
        momlevel=0.5*mindism0
    ax_moment0.contour(moment0[0].data, transform=ax_moment0.get_transform(im_wcs),
               levels=momlevel, colors='white',linewidths=1.5*size_factor , zorder =4)
    ax_moment0.contour(moment0[0].data, transform=ax_moment0.get_transform(im_wcs),
              levels=momlevel, colors='k',zorder=6, linewidths=1.2*size_factor)
    ax_moment0.contour(moment0_mod[0].data, transform=ax_moment0.get_transform(im_wcs),
               levels=momlevel, colors='white',linewidths=1.2*size_factor , zorder =7)
    ax_moment0.contour(moment0_mod[0].data, transform=ax_moment0.get_transform(im_wcs),
              levels=momlevel, colors='r',zorder=8, linewidths=0.9*size_factor)
    xmin, xmax = ax_moment0.get_xlim()
    ymin, ymax = ax_moment0.get_ylim()
    if xmax > ymax:
        diff = int(xmax-ymax)/2.
        ax_moment0.set_ylim(ymin-diff,ymax+diff)
        ymin, ymax = ax_moment0.get_ylim()
    else:
        diff = int(ymax-xmax)/2.
        ax_moment0.set_xlim(xmin-diff,xmax+diff)
        xmin, xmax = ax_moment0.get_xlim()
    ghxloc, ghyloc = im_wcs.wcs_pix2world(float(xmin+(xmax-xmin)/18.), float(ymin+(ymax-ymin)/18.), 1.)
    localoc = [float(ghxloc),float(ghyloc) ]
    widthb = moment0[0].header['BMIN']
    heightb = moment0[0].header['BMAJ']
    try:
        angleb  = moment0[0].header['BPA']
    except KeyError:
        angleb = 0.
    beam = Ellipse(xy=localoc, width=widthb, height=heightb, angle=angleb, transform=ax_moment0.get_transform('fk4'),
           edgecolor='k', lw=1, facecolor='none', hatch='/////',zorder=15)
    ax_moment0.add_patch(beam)
    # colorbar
    divider = make_axes_locatable(ax_moment0)
    cax = divider.append_axes("top", size="5%", pad=0.05, axes_class=maxes.Axes)
    cbar = plt.colorbar(moment0_plot, cax=cax, orientation='horizontal')
    cax.xaxis.set_ticks_position('top')
    cbar.set_ticks([min_color, max_color])
    cbar.ax.set_title(f"{moment0[0].header['BUNIT']}", y= 0.2*size_factor**2)

    '''the channel width should be 1 as the map is already in Jy/beam*km/s'''
    column_levels = sf.columndensity(Configuration,momlevel*1000.,\
        systemic = FAT_Model[Vars_to_plot.index('VSYS'),0], channel_width=1)

    if 1e21 < np.min(column_levels):
        fact= 1e21
        str_fact = r'$\times 10^{21} {\rm cm}^{-2}$'
    elif 1e20 < np.min(column_levels) < 1e21:
        fact= 1e20
        str_fact = r'$\times 10^{20} {\rm cm}^{-2}$'
    elif 1e19 < np.min(column_levels) < 1e20:
        fact= 1e19
        str_fact = r'$\times 10^{19} {\rm cm}^{-2}$'
    elif 1e18 < np.min(column_levels) < 1e19:
        fact= 1e18
        str_fact = r'$\times 10^{18} {\rm cm}^{-2}$'
    else:
        fact= 1e17
        str_fact = r'$\times 10^{17} {\rm cm}^{-2}$'
    if len(momlevel) < 4:
        info_string = f"The contours are at {', '.join(['{:.1f}'.format(x/fact) for x in column_levels])} {str_fact}."
    else:
        info_string = f"The contours are at {', '.join(['{:.1f}'.format(x/fact) for x in column_levels[0:4]])}"
        counter = 4
        while counter < len(column_levels):
            info_string = info_string+f"\n {', '.join(['{:.1f}'.format(x/fact) for x in column_levels[counter:counter+7]])}"
            counter += 7
        info_string = info_string+f" {str_fact}."
    #info_string = f"The contours are at {column_levels}."
    ax_moment0.text(-0.1,-0.2,info_string, va='top',ha='left', color='black',transform = ax_moment0.transAxes,
              bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7)
    # No further need for the moment maps
    moment0.close()
    moment0_mod.close()
#-----------------------------------------------------------------Velocity Field------------------------------------------------------
    ax_moment1 = Overview.add_subplot(gs[10:16, 0:6], projection=im_wcs)
    ax_moment1.set_label('Velocity Field')
    #Comp_ax1.set_facecolor('black')
    # we need contour levels and
    inclination_correction = sf.set_limits(FAT_Model[Vars_to_plot.index('INCL'),0]+12.5,20.,90.)
    velocity_width= sf.set_limits(1.25*np.nanmax(FAT_Model[Vars_to_plot.index('VROT'),:])*np.sin(np.radians(inclination_correction)),30.,700.)

    max_color= FAT_Model[Vars_to_plot.index('VSYS'),0]+velocity_width
    min_color= FAT_Model[Vars_to_plot.index('VSYS'),0]-velocity_width

    moment1_plot = ax_moment1.imshow(moment1[0].data, cmap='rainbow', origin='lower', alpha=1, vmin = min_color, vmax = max_color )
    plt.ylabel('DEC (J2000)')
    #Stupid python suddenly finds its own labels
    plt.xlabel('RA J2000')

    # contours
    velocity_step=sf.set_limits(int((int(max_color-min_color)*0.9)/20.),1.,30.)
    integer_array = np.linspace(0,20,21)-10
    momlevel = [FAT_Model[Vars_to_plot.index('VSYS'),0]+x*velocity_step for x in integer_array if min_color < FAT_Model[Vars_to_plot.index('VSYS'),0]+x*velocity_step < max_color]
    ax_moment1.contour(moment1[0].data, transform=ax_moment1.get_transform(im_wcs),
               levels=momlevel, colors='white',linewidths=1.5 *size_factor, zorder =4)
    ax_moment1.contour(moment1[0].data, transform=ax_moment1.get_transform(im_wcs),
              levels=momlevel, colors='k',zorder=6, linewidths=0.9*size_factor)
    ax_moment1.contour(moment1_mod[0].data, transform=ax_moment1.get_transform(im_wcs),
               levels=momlevel, colors='white',linewidths=1.2*size_factor , zorder =7)
    #ax_moment1.contour(moment1_mod[0].data, transform=ax_moment1.get_transform(im_wcs),
    #          levels=momlevel, colors='r',zorder=8, linewidths=0.9)
    xmin, xmax = ax_moment1.get_xlim()
    ymin, ymax = ax_moment1.get_ylim()
    if xmax > ymax:
        diff = int(xmax-ymax)/2.
        ax_moment1.set_ylim(ymin-diff,ymax+diff)
        ymin, ymax = ax_moment1.get_ylim()
        xmin, xmax = ax_moment1.get_xlim()
    else:
        diff = int(ymax-xmax)/2.
        ax_moment1.set_xlim(xmin-diff,xmax+diff)
    ghxloc, ghyloc = im_wcs.wcs_pix2world(float(xmin+(xmax-xmin)/18.), float(ymin+(ymax-ymin)/18.), 1.)
    localoc = [float(ghxloc),float(ghyloc) ]
    widthb = moment1[0].header['BMIN']
    heightb = moment1[0].header['BMAJ']
    try:
        angleb  = moment1[0].header['BPA']
    except KeyError:
        angleb = 0.
    beam = Ellipse(xy=localoc, width=widthb, height=heightb, angle=angleb, transform=ax_moment1.get_transform('fk4'),
           edgecolor='k', lw=1, facecolor='none', hatch='/////',zorder=15)
    ax_moment1.add_patch(beam)
    # colorbar
    divider = make_axes_locatable(ax_moment1)
    cax = divider.append_axes("top", size="5%", pad=0.05, axes_class=maxes.Axes)
    cbar = plt.colorbar(moment1_plot, cax=cax, orientation='horizontal')
    cax.xaxis.set_ticks_position('top')
    cbar.set_ticks([min_color, max_color])
    cbar.ax.set_title(f"{moment1[0].header['BUNIT']}", y= 0.2*size_factor**2)

    column_levels = ', '.join(["{:.1f}".format(x) for x in momlevel])
    info_string= f'''The contours start at {float(momlevel[0]):.1f} km/s'''
    if len(momlevel) > 1:
        info_string=f'''{info_string} \n and increase with {float(momlevel[1])-float(momlevel[0]):.1f} km/s.'''
    else:
        info_string=f'''{info_string}.'''


    ax_moment1.text(-0.1,-0.2,info_string, va='top',ha='left', color='black',transform = ax_moment1.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7)
    # No further need for the moment maps
    moment1.close()
    moment1_mod.close()
#-----------------------------------------------------------------Moment 2------------------------------------------------------
    ax_moment2 = Overview.add_subplot(gs[10:16:, 8:14], projection=im_wcs)
    ax_moment2.set_label('Moment2')
    #Comp_ax1.set_facecolor('black')
    # we need contour levels and
    max_color= sf.set_limits(np.nanmax(moment2[0].data),15,50)
    min_color= 0.

    moment2_plot = ax_moment2.imshow(moment2[0].data, cmap='rainbow' ,origin='lower', alpha=1, vmin = min_color, vmax = max_color )
    plt.ylabel('DEC (J2000)')
    #Stupid python suddenly finds its own labels
    plt.xlabel('RA J2000')

    # contours

    momlevel = np.linspace(min_color,max_color*0.8,5)

    ax_moment2.contour(moment2[0].data, transform=ax_moment2.get_transform(im_wcs),
               levels=momlevel, colors='white',linewidths=1.5*size_factor , zorder =4)
    ax_moment2.contour(moment2[0].data, transform=ax_moment2.get_transform(im_wcs),
              levels=momlevel, colors='k',zorder=6, linewidths=1.2*size_factor)
    ax_moment2.contour(moment2_mod[0].data, transform=ax_moment2.get_transform(im_wcs),
               levels=momlevel, colors='white',linewidths=1.2*size_factor , zorder =7)

    xmin, xmax = ax_moment2.get_xlim()
    ymin, ymax = ax_moment2.get_ylim()
    if xmax > ymax:
        diff = int(xmax-ymax)/2.
        ax_moment2.set_ylim(ymin-diff,ymax+diff)
        ymin, ymax = ax_moment2.get_ylim()
    else:
        diff = int(ymax-xmax)/2.
        ax_moment2.set_xlim(xmin-diff,xmax+diff)
        xmin, xmax = ax_moment2.get_xlim()

    ghxloc, ghyloc = im_wcs.wcs_pix2world(float(xmin+(xmax-xmin)/18.), float(ymin+(ymax-ymin)/18.), 1.)
    localoc = [float(ghxloc),float(ghyloc) ]
    widthb = moment2[0].header['BMIN']
    heightb = moment2[0].header['BMAJ']
    try:
        angleb  = moment2[0].header['BPA']
    except KeyError:
        angleb = 0.

    beam = Ellipse(xy=localoc, width=widthb, height=heightb, angle=angleb, transform = ax_moment2.get_transform('fk4'),
           edgecolor='k', lw=1, facecolor='none', hatch='/////',zorder=15)
    ax_moment2.add_patch(beam)


    # colorbar
    divider = make_axes_locatable(ax_moment2)
    cax = divider.append_axes("top", size="5%", pad=0.05, axes_class=maxes.Axes)
    cbar = plt.colorbar(moment2_plot, cax=cax, orientation='horizontal')
    cax.xaxis.set_ticks_position('top')
    cbar.set_ticks([min_color, max_color])
    #cbar.set_title(label=f"{moment2[0].header['BUNIT']}")
    cbar.ax.set_title(f"{moment2[0].header['BUNIT']}", y= 0.2*size_factor**2)

    column_levels = ', '.join(["{:.1f}".format(x) for x in momlevel])
    if len(momlevel) < 4:
        info_string = f"The contours are at {column_levels} km/s."
    else:
        info_string = f"The contours are at {', '.join(['{:.1f}'.format(x) for x in momlevel[0:4]])}"
        counter = 4
        while counter < len(momlevel):
            info_string = info_string+f"\n {', '.join(['{:.1f}'.format(x) for x in momlevel[counter:counter+7]])}"
            counter += 7
            info_string = info_string+" km/s."

    ax_moment2.text(-0.1,-0.2,info_string, va='top',ha='left', color='black',transform = ax_moment2.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7)
    # No further need for the moment maps
    moment2.close()
    moment2_mod.close()



#__________________------------------------------------------------------------PV Diagram

    extract_angle = np.mean(FAT_Model[Vars_to_plot.index('PA'),0:round(len(FAT_Model[Vars_to_plot.index('PA'),:])/2.)])
    PV = extract_pv(Configuration,cube,extract_angle, \
                    center = [float(FAT_Model[Vars_to_plot.index('XPOS'),0]),float(FAT_Model[Vars_to_plot.index('YPOS'),0]),float(FAT_Model[Vars_to_plot.index('VSYS'),0]*1000.)], \
                    convert=1000.)
    if not os.path.exists(f"{Configuration['FITTING_DIR']}/Finalmodel/{Configuration['BASE_NAME']}_final_xv.fits"):
        fits.writeto(f"{Configuration['FITTING_DIR']}/Finalmodel/{Configuration['BASE_NAME']}_final_xv.fits",PV[0].data,PV[0].header)
    PV_model = extract_pv(Configuration,cube_mod,extract_angle, \
                    center = [float(FAT_Model[Vars_to_plot.index('XPOS'),0]),float(FAT_Model[Vars_to_plot.index('YPOS'),0]),float(FAT_Model[Vars_to_plot.index('VSYS'),0]*1000.)], \
                    convert=1000.)
    if not os.path.exists(f"{Configuration['FITTING_DIR']}/Finalmodel/Finalmodel_xv.fits"):
        fits.writeto(f"{Configuration['FITTING_DIR']}/Finalmodel/Finalmodel_xv.fits",PV_model[0].data,PV_model[0].header)
    ratio=PV[0].header['NAXIS2']/PV[0].header['NAXIS1']
    # Then we want to plot our PV-Diagram
    '''
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        #As astropy is the dumbest piece of software ever invented and it cannot separate stray keywords from extra dimensions
        # we need to specify we want the first two axes.
        xv_proj = WCS(PV[0].header,naxis=[0,1])
        xv_model_proj = WCS(PV_model[0].header,naxis=[0,1])
    # As astropy is really a dumb piece of shit using a WCS messes up the the axes modifications
    #ax_PV = Overview.add_subplot(gs[2:8,8:14], projection=xv_proj)
    '''
    ax_PV = Overview.add_subplot(gs[2:8,8:14])
    #Comp_ax2.set_title('PV-Diagram')

    maxint= np.nanmax(PV[0].data)*0.85
    minint= np.nanmin(PV[0].data)/3.


    PV_plot = ax_PV.imshow(PV[0].data,  cmap='hot_r', origin='lower', alpha=1, vmin=minint, vmax=maxint,aspect='auto')

    xaxis = [PV[0].header['CRVAL1'] + (i - PV[0].header['CRPIX1'] + 1) * (PV[0].header['CDELT1']) for i in
             range(PV[0].header['NAXIS1'])]
    yaxis = [PV[0].header['CRVAL2'] + (i - PV[0].header['CRPIX2'] + 1) * (PV[0].header['CDELT2']) for i in
             range(PV[0].header['NAXIS2'])]
    #something is going wrong here, Fixed as usual astropy was fucking thing up
    ax_PV.set_xticks(range(len(xaxis))[0:-1:int(len(xaxis) / 5)])
    ax_PV.set_yticks(range(len(yaxis))[0:-1:int(len(yaxis) / 5)])
    ax_PV.set_xticklabels(['{:10.1f}'.format(i) for i in xaxis[0:-1:int(len(xaxis) / 5)]])
    ax_PV.set_yticklabels(['{:10.1f}'.format(i) for i in yaxis[0:-1:int(len(yaxis) / 5)]])

    #Add some contours
    neg_cont = np.array([-3,-1.5],dtype=float)*Configuration['NOISE']
    if  np.nanmax(PV[0].data) * 0.95 < 96*Configuration['NOISE']:
        pos_cont =  np.array([1.5,3.,6,12,24,48,96],dtype=float)*Configuration['NOISE']
    else:
        pos_cont =  np.array([0,1,2,3,4,5,6,7],dtype=float)*(np.nanmax(PV[0].data) * 0.95-3.*Configuration['NOISE'])/7. +3.*Configuration['NOISE']
    pos_cont = np.array([x for x in pos_cont if x <= np.nanmax(PV[0].data) * 0.95],dtype=float)
    sf.print_log(f'''MAKE_OVERVIEW_PLOT: postive {pos_cont}, negative {neg_cont}, noise {Configuration['NOISE']}
''',Configuration,case=['debug_add'] )
    if pos_cont.size == 0:
        pos_cont = 0.5 * np.nanmax(PV[0].data) * 0.95

    #ax_PV.contour(PV[0].data, levels=pos_cont, colors='k',transform=ax_PV.get_transform(xv_proj))
    #ax_PV.contour(PV[0].data, levels=neg_cont, colors='grey',linestyles='--',transform=ax_PV.get_transform(xv_proj))
    #ax_PV.contour(PV_model[0].data, levels=pos_cont, colors='b',transform=ax_PV.get_transform(xv_model_proj),linewidths=1.)
    ax_PV.contour(PV[0].data, levels=pos_cont,linewidths=1.*size_factor, colors='k')
    ax_PV.contour(PV[0].data, levels=neg_cont,linewidths=1.*size_factor, colors='grey',linestyles='--')
    ax_PV.contour(PV_model[0].data, levels=pos_cont, colors='b',linewidths=1.*size_factor)

    momlevel = np.hstack((neg_cont,pos_cont))
    column_levels = ', '.join(["{:.1f}".format(x*1000.) for x in momlevel])
    if len(momlevel) < 4:
        info_string = f"The contours are at {column_levels} mJy/beam"
    else:
        info_string = f"The contours are at {', '.join(['{:.1f}'.format(x*1000.) for x in momlevel[0:4]])}"
        counter = 5
        while counter < len(momlevel):
            info_string = info_string+f"\n {', '.join(['{:.1f}'.format(x*1000.) for x in momlevel[counter:counter+7]])}"
            counter += 7
        info_string = info_string+" mJy/beam."

    divider = make_axes_locatable(ax_PV)
    cax = divider.append_axes("top", size="5%", pad=0.05, axes_class=maxes.Axes)
    cbar = plt.colorbar(PV_plot, cax=cax, orientation='horizontal')
    cax.xaxis.set_ticks_position('top')
    cbar.set_ticks([minint, maxint])
    cbar.ax.set_title(f"{PV[0].header['BUNIT']}", y= 0.2*size_factor**2)

    ax_PV.text(-0.1,-0.2,info_string, va='top',ha='left', color='black',transform = ax_PV.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad= 0.,alpha=0.),zorder=7)

    #cf.plot_fits(filename, Comp_ax2, cmap='hot_r', aspect=ratio, cbar ='horizontal')
    ax_PV.set_xlabel("Offset (arcsec)")
    ax_PV.set_ylabel("Velocity (km s$^{-1}$)")
    PV.close()
    PV_model.close()

# ADD a legend for the plots

    ax_legend = Overview.add_subplot(gs[18:27,0:2],frameon = False)
    plt.tick_params(
            axis='both',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            direction = 'in',
            bottom=False,      # ticks along the bottom edge are off
            top=False,         # ticks along the top edge are off
            labelbottom=False,
            right = False,
            left= False,
            labelleft = False)
    legend = ['Appr.','Rec.','Correct Appr.','Correct Rec.','Unsm. Appr.','Unsm. Rec.']
    plt.plot([0,1],[0,1],'k',marker='o',label=legend[0])
    plt.plot([0,1],[0,1],'r',marker='o',label=legend[1])
    if len(Input_Model) > 0:
        plt.plot([0,1],[0,1],'b',marker='o',label=legend[2])
        plt.plot([0,1],[0,1],'yellow',marker='o',label=legend[3])
    if len(Extra_Model) > 0:
        plt.plot([0,1],[0,1],'k',marker='o',label=legend[4],alpha=0.2)
        plt.plot([0,1],[0,1],'r',marker='o',label=legend[5],alpha=0.2)
    plt.plot([0,1],[0,1],'white',marker='o',linewidth=5)
    plt.ylim(0.4,0.6)
    plt.xlim(0.4,0.6)

    chartBox = ax_legend.get_position()
    ax_legend.set_position([chartBox.x0, chartBox.y0, chartBox.width*1.0, chartBox.height])
    ax_legend.legend(loc='upper left', bbox_to_anchor = (-1.0,1.03),shadow=True, ncol=1)


# ------------------------------Rotation curves------------------------------------

    labelfont= {'family':font_name,
            'weight':'normal',
            'size':10*size_factor}
    ax_RC = plot_parameters(Configuration,Vars_to_plot,FAT_Model,gs[19:22,3:9],\
                            Overview,'VROT',Input_Model = Input_Model,initial_extent= sof_basic_extent[0], \
                            initial = sof_basic_maxrot[0],Extra_Model = Extra_Model)
    ymin =np.min([FAT_Model[Vars_to_plot.index('VROT'),1:],FAT_Model[Vars_to_plot.index('VROT_2'),1:]])
    if len(Extra_Model) > 0:
        ymin2 =np.min([Extra_Model[Vars_to_plot.index('VROT'),1:],Extra_Model[Vars_to_plot.index('VROT_2'),1:]])
        ymax2 =np.max([Extra_Model[Vars_to_plot.index('VROT'),1:],Extra_Model[Vars_to_plot.index('VROT_2'),1:]])
    else:
        ymin2 = ymin
        ymax2 = ymax
    ymax =np.max([FAT_Model[Vars_to_plot.index('VROT'),1:],FAT_Model[Vars_to_plot.index('VROT_2'),1:]])

    if len(Input_Model) > 0:
        ymin3 =np.min([Input_Model[Vars_to_plot.index('VROT'),1:],Input_Model[Vars_to_plot.index('VROT_2'),1:]])
        ymax3 =np.max([Input_Model[Vars_to_plot.index('VROT'),1:],Input_Model[Vars_to_plot.index('VROT_2'),1:]])
    else:
        ymin3=ymin
        ymax3= ymax
    ymin = np.min([ymin,ymin2,ymin3])
    ymax = np.max([ymax,ymax2,ymax3])

    buffer = np.mean([ymin,ymax])/20.
    ymin= ymin-buffer
    ymax = ymax+buffer

    ax_RC.set_ylim(ymin,ymax)

    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        direction = 'in',
        bottom=True,      # ticks along the bottom edge are off
        top=True,         # ticks along the top edge are off
        labelbottom=False)
    plt.ylabel('RC (km s$^{-1}$)',**labelfont)
    arcmin,arcmax = ax_RC.get_xlim()
    sec_ax = ax_RC.twiny()
    sec_ax.set_xlim(sf.convertskyangle(Configuration,arcmin),sf.convertskyangle(Configuration,arcmax))
    sec_ax.figure.canvas.draw()
    sec_ax.set_xlabel('Radius (kpc)',va='bottom',**labelfont)

# ------------------------------Inclination------------------------------------
    ax_INCL = plot_parameters(Configuration,Vars_to_plot,FAT_Model,gs[22:25,3:9],\
                              Overview,'INCL',Input_Model = Input_Model,initial_extent= sof_basic_extent[0], \
                              initial =sof_basic_inclination[0],Extra_Model = Extra_Model  )

    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        direction = 'in',
        bottom=True,      # ticks along the bottom edge are off
        top=True,         # ticks along the top edge are off
        labelbottom=False)
    if 'INCL' in Configuration['FIXED_PARAMETERS'][0]:
        ax_INCL.text(1.01,0.5,'Forced Flat', rotation =-90,va='center',ha='left', color='black',transform = ax_INCL.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7,fontsize=10*size_factor)
    plt.ylabel('Incl ($^{\circ}$)',**labelfont)
    arcmin,arcmax = ax_INCL.get_xlim()
    sec_ax = ax_INCL.twiny()
    sec_ax.set_xlim(sf.convertskyangle(Configuration,arcmin),sf.convertskyangle(Configuration,arcmax))
    sec_ax.set_xticklabels([])
    sec_ax.figure.canvas.draw()

# ------------------------------PA------------------------------------
    ax_PA = plot_parameters(Configuration,Vars_to_plot,FAT_Model,gs[25:28,3:9],\
                            Overview,'PA',Input_Model = Input_Model,initial_extent= sof_basic_extent[0],\
                            initial = sof_basic_pa[0],Extra_Model = Extra_Model )

    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        direction = 'in',
        bottom=True,      # ticks along the bottom edge are off
        top=True,         # ticks along the top edge are off
        labelbottom=True)
    if 'PA' in Configuration['FIXED_PARAMETERS'][0]:
        ax_PA.text(1.01,0.5,'Forced Flat', va='center',ha='left', color='black',rotation = -90, transform = ax_PA.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7,fontsize=10*size_factor)
    plt.xlabel('Radius (arcsec)',**labelfont)
    plt.ylabel('PA ($^{\circ}$)',**labelfont)
    arcmin,arcmax = ax_PA.get_xlim()
    sec_ax = ax_PA.twiny()
    sec_ax.set_xlim(sf.convertskyangle(Configuration,arcmin),sf.convertskyangle(Configuration,arcmax))
    sec_ax.set_xticklabels([])
    sec_ax.figure.canvas.draw()

# ------------------------------SDIS------------------------------------
    ax_SDIS = plot_parameters(Configuration,Vars_to_plot,FAT_Model,gs[19:22,12:18],\
                              Overview,'SDIS',Input_Model = Input_Model,initial_extent= sof_basic_extent[0],\
                              Extra_Model = Extra_Model,initial = 8. )
    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        direction = 'in',
        bottom=True,      # ticks along the bottom edge are off
        top=True,         # ticks along the top edge are off
        labelbottom=False,
        labeltop= True)
    if 'SDIS' in Configuration['FIXED_PARAMETERS'][0]:
        ax_SDIS.text(1.01,0.5,'Forced Flat',rotation=-90, va='center',ha='left', color='black',transform = ax_SDIS.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7,fontsize=10*size_factor)

    plt.ylabel('Disp (km s$^{-1}$)',**labelfont)
    arcmin,arcmax = ax_SDIS.get_xlim()
    sec_ax = ax_SDIS.twiny()
    sec_ax.set_xlim(sf.convertskyangle(Configuration,arcmin),sf.convertskyangle(Configuration,arcmax))
    sec_ax.figure.canvas.draw()
    sec_ax.set_xlabel('Radius (kpc)',va='bottom',**labelfont)


# ------------------------------Scale height------------------------------------
    ax_Z0 = plot_parameters(Configuration,Vars_to_plot,FAT_Model,gs[22:25,12:18],\
                            Overview,'Z0',Input_Model = Input_Model,initial_extent= sof_basic_extent[0],\
                            Extra_Model = Extra_Model,initial = sf.convertskyangle(Configuration,0.2,physical=True) )

    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        direction = 'in',
        bottom=True,      # ticks along the bottom edge are off
        top=True,         # ticks along the top edge are off
        labelbottom=False,
        labeltop=False)
    if 'Z0' in Configuration['FIXED_PARAMETERS'][0]:
        ax_Z0.text(1.25,0.5,'Forced Flat',rotation=-90, va='center',ha='left', color='black',transform = ax_Z0.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7,fontsize=10*size_factor)
    plt.ylabel('Z0 (arcsec)',**labelfont)
    arcmin,arcmax = ax_Z0.get_ylim()
    sec_ax = ax_Z0.twinx()
    sec_ax.set_ylim(sf.convertskyangle(Configuration,arcmin),sf.convertskyangle(Configuration,arcmax))
    sec_ax.figure.canvas.draw()
    sec_ax.set_ylabel('Z0 (kpc)',rotation=-90,va='bottom',**labelfont)
    arcmin,arcmax = ax_Z0.get_xlim()
    sec_ax = ax_Z0.twiny()
    sec_ax.set_xticklabels([])
    sec_ax.set_xlim(sf.convertskyangle(Configuration,arcmin),sf.convertskyangle(Configuration,arcmax))
    sec_ax.figure.canvas.draw()


# ------------------------------SBR------------------------------------
    ax_SBR = plot_parameters(Configuration,Vars_to_plot,FAT_Model,gs[25:28,12:18],\
                             Overview,'SBR',Input_Model = Input_Model,initial_extent= sof_basic_extent[0],\
                             Extra_Model = Extra_Model )

    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        direction = 'in',
        bottom=True,      # ticks along the bottom edge are off
        top=True,         # ticks along the top edge are off
        labeltop=False)


    plt.ylabel('SBR \n (Jy km s$^{-1}$ arcsec$^{-2}$)',**labelfont)
    plt.xlabel('Radius (arcsec)',**labelfont)

    jymin,jymax = ax_SBR.get_ylim()
    sec_ax = ax_SBR.twinx()
    sec_ax.set_ylim(sf.columndensity(Configuration,jymin*1000.,arcsquare = True)/1e20,sf.columndensity(Configuration,jymax*1000.,arcsquare = True)/1e20)
    sec_ax.set_ylabel('Col. Dens. \n (x10$^{20}$ cm$^{-2}$)',rotation=-90,va='bottom',**labelfont)
    sec_ax.figure.canvas.draw()
    if 'SBR' in Configuration['FIXED_PARAMETERS'][0]:
        ax_SBR.text(1.25,0.5,'Forced Gaussian',rotation=-90, va='center',ha='left', color='black',transform = ax_SBR.transAxes,
          bbox=dict(facecolor='white',edgecolor='white',pad=0.,alpha=0.),zorder=7,fontsize=10*size_factor)
    sec_ax = ax_SBR.twiny()
    arcmin,arcmax = ax_SBR.get_xlim()
    sec_ax.set_xticklabels([])
    sec_ax.set_xlim(sf.convertskyangle(Configuration,arcmin),sf.convertskyangle(Configuration,arcmax))
    sec_ax.figure.canvas.draw()


#----------------------------------------------Distance vs VSYS -----------------------------------------
    ax_VSYS = Overview.add_subplot(gs[2:6,16:20])
    plt.xlabel('Sys. Vel. (km s$^{-1}$)',**labelfont)
    plt.ylabel('Distance (Mpc)',**labelfont)

    plt.errorbar(float(FAT_Model[Vars_to_plot.index('VSYS'),0]),float(Configuration['DISTANCE'])\
                ,xerr=float(FAT_Model[Vars_to_plot.index('# VSYS_ERR'),0]), c='k',zorder= 3,fmt="o")
    #plt.scatter(float(FAT_Model[0,Vars_to_plot.index('VSYS')]),float(Configuration['DISTANCE']),c='k',zorder= 3)
    if len(Extra_Model) > 0:
        plt.scatter(float(Extra_Model[Vars_to_plot.index('VSYS'),0]),float(Configuration['DISTANCE']),c='r',alpha = 0.5,zorder=1)
    if len(Input_Model) > 0:
        plt.scatter(float(Input_Model[Vars_to_plot.index('VSYS'),0]),float(Configuration['DISTANCE']),c='b',zorder= 2)
    plt.scatter(sof_basic_vsys[0],float(Configuration['DISTANCE']),marker='x',alpha=0.5, c = 'k')
    xmin,xmax = ax_VSYS.get_xlim()
    ymin,ymax = ax_VSYS.get_ylim()

    #plt.scatter(sof_basic_vsys[0],float(Configuration['DISTANCE']),marker='x',alpha=0.5, c = 'k')
    vall= np.linspace(0,15000,100)
    Distanc=vall/70.
    plt.plot(vall,Distanc,'k--',alpha=0.5)
    ax_VSYS.set_xlim(xmin, xmax)
    left = float(FAT_Model[Vars_to_plot.index('VSYS'),0])-cube[0].header['CDELT3']/2000.
    bottom = ymin-5
    width =  cube[0].header['CDELT3']/1000.
    height = ymax-ymin+50
#plt.fill([sof_basic_vsys[0]-Cube[0].header['CDELT3']/2000.,sof_basic_vsys[0]+Cube[0].header['CDELT3']/2000.],[ymin-50,ymax+50],c='k',alpha=0.5)
    why = plt.Rectangle((left,bottom), width,height,facecolor='black',alpha=0.4 , zorder=1)
    ax_VSYS.add_patch(why)
    ax_VSYS.set_ylim(ymin, ymax)
#----------------------------------------------RA vs DEC -----------------------------------------
    ax_RAD = Overview.add_subplot(gs[8:12,16:20])

    plt.errorbar(float(FAT_Model[Vars_to_plot.index('XPOS'),0]),\
                 float(FAT_Model[Vars_to_plot.index('YPOS'),0]),
                xerr=float(FAT_Model[Vars_to_plot.index('# XPOS_ERR'),0]),\
                yerr=float(FAT_Model[Vars_to_plot.index('# YPOS_ERR'),0]), c='k',zorder= 3,fmt="o")

    plt.scatter(float(FAT_Model[Vars_to_plot.index('XPOS'),0]),float(FAT_Model[Vars_to_plot.index('YPOS'),0]),c='k',zorder=3,label = 'Final')
    if len(Extra_Model) > 0:
        lab = 'Unsmoothed'
        alpha =0.5
        plt.scatter(float(Extra_Model[Vars_to_plot.index('XPOS'),0]),float(Extra_Model[Vars_to_plot.index('YPOS'),0]),c='r',zorder=1,alpha=alpha,label=lab)
    if len(Input_Model) > 0:
        plt.scatter(float(Input_Model[Vars_to_plot.index('XPOS'),0]),float(Input_Model[Vars_to_plot.index('YPOS'),0]),c='b',zorder =2,label='Input')
    plt.scatter(sof_basic_ra[0],sof_basic_dec[0],marker='x',alpha=0.5, c = 'k',label='Initial')
    mod_ell = Ellipse(xy=[float(FAT_Model[Vars_to_plot.index('XPOS'),0]),float(FAT_Model[Vars_to_plot.index('YPOS'),0])], width=cube[0].header['BMAJ'] , height=cube[0].header['BMAJ'], angle=0,
               edgecolor='none', alpha=0.4, lw=4, facecolor='k', hatch = None, zorder=-1)
    ax_RAD.add_patch(mod_ell)
    ax_RAD.legend(loc='upper left', bbox_to_anchor=(0.0, -0.3), shadow=True, ncol=1)
    plt.xlabel('RA ($^{\circ}$)',**labelfont)
    plt.ylabel('DEC ($^{\circ}$)',**labelfont)
    cube_mod.close()
    cube.close()
    channels_map.close()
    plt.savefig(f"{Configuration['FITTING_DIR']}Overview.png", bbox_inches='tight')
    #plt.savefig(f"Overview_Test.png", bbox_inches='tight')
    plt.close()

make_overview_plot.__doc__ =f'''
 NAME:
    make_overview_plot(Configuration,Fits_Files ):

 PURPOSE:
    Create a plot that shows the various stages of the fitting and output in a handy overview

 CATEGORY:
    write_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Locations of the Fits files used by FAT

 OPTIONAL INPUTS:


 OUTPUTS:
    The overview plot

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def plot_parameters(Configuration,Vars_to_plot,FAT_Model,location,Figure,parameter,\
                    Input_Model = [],legend = ['Empty','Empty','Empty','Empty'],
                    initial = None, initial_extent=  None, Extra_Model = [] ):
    sf.print_log(f'''PLOT_PARAMETERS: We are starting to plot {parameter}
''', Configuration,case=['debug_start'])
    ax = Figure.add_subplot(location)
    try:
        yerr = FAT_Model[Vars_to_plot.index(f'# {parameter}_ERR'),:]
    except ValueError:
        yerr =np.zeros(len(FAT_Model[Vars_to_plot.index('RADI'),:]))
    sf.print_log(f'''PLOT_PARAMETERS: We found these errors {yerr}
''', Configuration,case=['debug_add'])

    ax.errorbar(FAT_Model[Vars_to_plot.index('RADI'),:],FAT_Model[Vars_to_plot.index(f'{parameter}'),:],yerr= yerr, c ='k', label=f'{legend[0]}',zorder=3)
    ax.plot(FAT_Model[Vars_to_plot.index('RADI'),:],FAT_Model[Vars_to_plot.index(f'{parameter}'),:],'ko', ms = 3.,zorder=3)
    if np.sum(FAT_Model[Vars_to_plot.index(f'{parameter}_2'),:]) != 0.:
        diff = np.sum(abs(FAT_Model[Vars_to_plot.index(f'{parameter}_2'),:]-FAT_Model[Vars_to_plot.index(f'{parameter}'),:]))
        if diff != 0.:
            try:
                yerr = FAT_Model[Vars_to_plot.index(f'# {parameter}_2_ERR'),:]
            except ValueError:
                yerr =np.zeros(len(FAT_Model[Vars_to_plot.index('RADI'),:]))
            ax.errorbar(FAT_Model[Vars_to_plot.index('RADI'),:],FAT_Model[Vars_to_plot.index(f'{parameter}_2'),:],yerr= yerr, c ='r', label=f'{legend[1]}',zorder=3)
            ax.plot(FAT_Model[Vars_to_plot.index('RADI'),:],FAT_Model[Vars_to_plot.index(f'{parameter}_2'),:],'ro', ms = 3.,zorder=3)

    if len(Input_Model) > 0:
        if np.sum(Input_Model[Vars_to_plot.index(f'{parameter}'),:]) != 0.:
            last_index = int(np.where(Input_Model[Vars_to_plot.index(f'{parameter}'),:] != 0.)[0][-1])
            try:
                yerr = Input_Model[Vars_to_plot.index(f'# {parameter}_ERR'),:last_index]
            except ValueError:
                yerr =np.zeros(len(Input_Model[Vars_to_plot.index('RADI'),:last_index]))
            ax.errorbar(Input_Model[Vars_to_plot.index('RADI'),:last_index],Input_Model[Vars_to_plot.index(f'{parameter}'),:last_index],yerr= yerr, c ='b',linestyle='-', label=f'{legend[2]}',zorder=2)
            ax.plot(Input_Model[Vars_to_plot.index('RADI'),:last_index],Input_Model[Vars_to_plot.index(f'{parameter}'),:last_index],'bo',linestyle='-', ms = 3.,zorder=2)
        if np.sum(Input_Model[Vars_to_plot.index(f'{parameter}_2'),:]) != 0.:
            diff = np.sum(abs(Input_Model[Vars_to_plot.index(f'{parameter}_2'),:]-Input_Model[Vars_to_plot.index(f'{parameter}'),:]))
            if diff != 0.:
                last_index = int(np.where(Input_Model[Vars_to_plot.index(f'{parameter}_2'),:] != 0.)[0][-1])
                try: ###### Keep swapping indices
                    yerr = Input_Model[Vars_to_plot.index(f'# {parameter}_2_ERR'),:last_index]
                except ValueError:
                    yerr =np.zeros(len(Input_Model[Vars_to_plot.index('RADI'),:last_index]))

                ax.errorbar(Input_Model[Vars_to_plot.index('RADI'),:last_index],Input_Model[Vars_to_plot.index(f'{parameter}_2'),:last_index],yerr= yerr, c ='yellow', label=f'{legend[3]}',zorder=2)
                ax.plot(Input_Model[Vars_to_plot.index('RADI'),:last_index,],Input_Model[Vars_to_plot.index(f'{parameter}_2'),:last_index],'yellow',zorder=2,marker ='o',linestyle='-' , ms = 3.)
    ymin,ymax = ax.get_ylim()
    if len(Extra_Model) > 0:
        ax.plot(Extra_Model[Vars_to_plot.index('RADI'),:],Extra_Model[Vars_to_plot.index(f'{parameter}'),:],'ko',linestyle='-', ms = 3., alpha=0.2,zorder=1)
        if np.sum(Extra_Model[Vars_to_plot.index(f'{parameter}_2'),:]) != 0.:
            diff = np.sum(abs(Extra_Model[Vars_to_plot.index(f'{parameter}_2'),:]-Extra_Model[Vars_to_plot.index(f'{parameter}'),:]))
            if diff != 0.:
                ax.plot(Extra_Model[Vars_to_plot.index('RADI'),:],Extra_Model[Vars_to_plot.index(f'{parameter}_2'),:],'r', alpha=0.2,zorder=1,marker ='o',linestyle='-' , ms = 3.)
    xmin,xmax = ax.get_xlim()
    if initial:
        ax.plot([xmin-5,xmax+5],[float(initial),float(initial)],c='k',alpha=0.4, linestyle ='--')
    if initial_extent:
        ax.plot([float(initial_extent)/2.,float(initial_extent)/2.],[ymin-5,ymax+5],c='k',alpha=0.4, linestyle ='--')
    ax.set_xlim(xmin,xmax)
    ax.set_ylim(ymin,ymax)
    return ax

plot_parameters.__doc__ =f'''
     NAME:
        plot_parameters

     PURPOSE:
        Sub plot a specific parameter from the fit

     CATEGORY:
        write_functions

     INPUTS:
        Configuration = Standard FAT configuration
        Vars_to_plot =List with order of the variables in the FAT_Model
        FAT_Model = Array with the values for all parameters
        location = location in the figure where to put this subplot
        Figure = the figure in which we are plotting
        parameter = the parameter to plot

     OPTIONAL INPUTS:


        Input_Model = []
        A model to compare to, should be ordered the same as the FAT_Model

        legend = ['Empty','Empty','Empty','Empty']
        legend names for fat_model and Input model

        initial = 'No Value'
        Initial guess of the parameter

        Extra_Model = []
        A third model to plot weakly behing the main model

     OUTPUTS:
        ax object from matplot lib

     OPTIONAL OUTPUTS:

     PROCEDURES CALLED:
        Unspecified

     NOTE:
'''



def plot_usage_stats(Configuration ):
    with open(f"{Configuration['LOG_DIRECTORY']}Usage_Statistics.txt") as file:
        lines = file.readlines()

    labels = {'FAT': {'label':[], 'Time':[]}, 'Tirific':{'label':[], 'Time':[]}}
    loads = {'FAT':{'CPU':[],'MEM':[],'Time':[]},'Tirific':{'CPU':[],'MEM':[],'Time': []}}
    try:
        mpl_fm.fontManager.addfont(Configuration['FONT_FILE'])
        font_name = mpl_fm.FontProperties(fname=Configuration['FONT_FILE']).get_name()
    except FileNotFoundError:
        font_name = 'DejaVu Sans'
    labelfont = {'family': font_name,
             'weight': 'normal',
             'size': 4}
    current_stage = 'Not_Found'
    current_module = 'Unknown'
    startdate = 0
    for line in lines:
        line = line.strip()
        tmp = line.split(' ')
        if line[0] == '#':
            date = extract_date(f"{tmp[-2]} {tmp[-1]}")
        else:
            date = extract_date(f"{tmp[0]} {tmp[1]}")
        if startdate == 0:
            startdate = date
        diff = date - startdate
        time = diff.total_seconds()/60.
        if line[0] == '#':
            if tmp[1] == 'TIRIFIC:':

                if tmp[2].lower() == 'initializing':
                    tmp2 = line.split('=')[1].split()
                    current_stage = tmp2[0].strip()
                    labels['Tirific']['label'].append(f'Initializing {current_stage}')
                    labels['Tirific']['Time'].append(time)
                elif tmp[2].lower() == 'finished':
                    labels['Tirific']['label'].append(f'Ended {current_stage}')
                    labels['Tirific']['Time'].append(time)
                    #current_stage = 'No Tirific'
                elif tmp[2].lower() == 'started':
                    labels['Tirific']['label'].append(f'Started {current_stage}')
                    labels['Tirific']['Time'].append(time)
            else:
                labels['FAT']['label'].append(f'Starting {tmp[1]}')
                labels['FAT']['Time'].append(time)
        else:
            if tmp[-1].lower() == 'tirific':
                loads['Tirific']['Time'].append(time)
                loads['Tirific']['CPU'].append(tmp[4])
                loads['Tirific']['MEM'].append(tmp[8])
            else:
                loads['FAT']['Time'].append(time)
                loads['FAT']['CPU'].append(tmp[4])
                loads['FAT']['MEM'].append(tmp[8])

    # Below thanks to P. Serra
    # Make single-PID figures and total figure
    #print(loads['Tirific']['Time'],loads['FAT']['Time'])
    if len(loads['Tirific']['Time']) > 0.:
        combined_time =  np.sort(np.array(loads['Tirific']['Time']+loads['FAT']['Time'],dtype=float))

        combined_loads ={'Tirific':{'CPU':np.interp(combined_time,np.array(loads['Tirific']['Time'],dtype=float),np.array(loads['Tirific']['CPU'],dtype=float)),\
                                    'MEM':np.interp(combined_time,np.array(loads['Tirific']['Time'],dtype=float),np.array(loads['Tirific']['MEM'],dtype=float))},\
                        'FAT':{'CPU':np.interp(combined_time,np.array(loads['FAT']['Time'],dtype=float),np.array(loads['FAT']['CPU'],dtype=float)),\
                                                    'MEM':np.interp(combined_time,np.array(loads['FAT']['Time'],dtype=float),np.array(loads['FAT']['MEM'],dtype=float))}

        }
        comb_list= labels['Tirific']['Time']+labels['FAT']['Time']
        comb_label = labels['Tirific']['label']+labels['FAT']['label']

        labels_times=np.array([x for x, _ in sorted(zip(comb_list, comb_label))],dtype=float)
        labels_comb = [x for _, x in sorted(zip(comb_list, comb_label))]

        fig, ax1 = plt.subplots(figsize = (8,6))
        fig.subplots_adjust(left = 0.1, right = 0.9, bottom = 0.3, top = 0.7)


        ax1.plot(combined_time,combined_loads['Tirific']['MEM'],'b-',lw=0.5)
        ax1.plot(combined_time,combined_loads['FAT']['MEM'],'b--',lw=0.5)
        ax1.set_ylim(0,np.max([combined_loads['Tirific']['MEM'],combined_loads['FAT']['MEM']]) \
                      +np.max([combined_loads['Tirific']['MEM'],combined_loads['FAT']['MEM']])/10.)
        ax1.set_ylabel('RAM (Mb) ', color='b')
        ax1.tick_params(axis='y', labelcolor='b')
        ax1.set_xlabel('time (min)', color='k',zorder=5)
        ax2 = ax1.twinx()
        ax2.plot(combined_time,combined_loads['Tirific']['CPU'],'r-',lw=0.5)
        ax2.plot(combined_time,combined_loads['FAT']['CPU'],'r--',lw=0.5)
        #ax2.plot(combined_time,combined_loads['FAT']['CPU'],'r--',lw=0.5)
        ax2.set_ylabel('CPUs (%)',color='r')
        ax2miny,ax2maxy = ax2.get_ylim()
        ax2.tick_params(axis='y', labelcolor='r')
        last_label = -100
        label_sep = combined_time[-1]/30.
        color, linest = '0.5', '--'
        labelfont = {'family': font_name,
                 'weight': 'normal',
                 'size': 6.5}
        prev_label = ''

        last_label_top = 0.
        last_label_bottom = 0.
        #for label,time in zip(labels['Tirific']['label'],labels['Tirific']['Time']):
        for label,time in zip(labels_comb,labels_times):

            if label in labels['Tirific']['label']:
                offset = 20.
                xoffset = 0.05
                vertical_start = ax2maxy
                va= 'bottom'
                ha= 'left'
                color = 'k'
                linest = '-'
            else:
                offset=-50
                xoffset = 0.025
                vertical_start = ax2miny
                va= 'top'
                ha='right'
                color = '0.5'
                linest = '--'


            if (prev_label == 'Initializing tmp_incl_check' or prev_label == 'Ended tmp_incl_check'):
                if (label != 'Initializing tmp_incl_check' and label != 'Ended tmp_incl_check') or \
                        time == labels_times[-1]    :

                    if time != labels_times[-1]:
                        ax2.axvline(x=prev_time, linestyle=linest, color=color, linewidth=0.05)
                        if label in labels['Tirific']['label']:
                            last_label = last_label_top = max(prev_time,last_label_top+label_sep)
                        else:
                            last_label = last_label_bottom = max(prev_time,last_label_bottom+label_sep)
                        ax2.text(last_label,vertical_start+offset,prev_label, va=va,ha=ha,rotation= 60, color='black',
                              bbox=dict(facecolor='white',edgecolor='white',pad= 0.,alpha=0.),zorder=7,fontdict = labelfont)
                        ax2.plot([prev_time,last_label+xoffset],[vertical_start,vertical_start+offset],linest,color=color,linewidth=0.05,clip_on=False)
                    ax2.axvline(x=time, linestyle=linest, color=color, linewidth=0.05)
                    if label in labels['Tirific']['label']:
                        last_label = last_label_top = max(time,last_label_top+label_sep)
                    else:
                        last_label = last_label_bottom = max(time,last_label_bottom+label_sep)
                    ax2.text(last_label,vertical_start+offset,label, va=va,ha=ha,rotation= 60, color='black',
                          bbox=dict(facecolor='white',edgecolor='white',pad= 0.,alpha=0.),zorder=7,fontdict = labelfont)
                    ax2.plot([time,last_label+xoffset],[vertical_start,vertical_start+offset],linest,color=color,linewidth=0.05,clip_on=False)
                else:
                    prev_time = time
            elif (prev_label == 'Initializing Error_Shaker' or prev_label == 'Ended Error_Shaker' or prev_label == 'Started Error_Shaker'):
                if (label != 'Initializing Error_Shaker' and label != 'Ended Error_Shaker' and  label != 'Started Error_Shaker') or \
                        time == labels_times[-1]:

                    #ax2.axvline(x=prev_time, linestyle=linest, color=color, linewidth=0.05)
                    #last_label = max(prev_time,last_label+label_sep)
                    #ax2.text(last_label,ax2maxy+20.,prev_label, va='bottom',ha='left',rotation= 60, color='black',
                    #      bbox=dict(facecolor='white',edgecolor='white',pad= 0.,alpha=0.),zorder=7,fontdict = labelfont)
                    #ax2.plot([prev_time,last_label+0.1],[ax2maxy,ax2maxy+20.],linest,color=color,linewidth=0.05,clip_on=False)
                    ax2.axvline(x=time, linestyle=linest, color=color, linewidth=0.05)
                    if label in labels['Tirific']['label']:
                        last_label = last_label_top = max(time,last_label_top+label_sep)
                    else:
                        last_label = last_label_bottom = max(time,last_label_bottom+label_sep)
                    ax2.text(last_label,vertical_start+offset,label, va=va,ha=ha,rotation= 60, color='black',
                          bbox=dict(facecolor='white',edgecolor='white',pad= 0.,alpha=0.),zorder=7,fontdict = labelfont)
                    ax2.plot([time,last_label+xoffset],[vertical_start,vertical_start+offset],linest,color=color,linewidth=0.05,clip_on=False)
                else:
                    prev_label = label
                    prev_time = time
            else:
                ax2.axvline(x=time, linestyle=linest, color=color, linewidth=0.05)
                if label in labels['Tirific']['label']:
                    last_label = last_label_top = max(time,last_label_top+label_sep)
                else:
                    last_label = last_label_bottom= max(time,last_label_bottom+label_sep)
                ax2.text(last_label,vertical_start+offset,label,va=va,ha=ha,rotation= 60, color='black',
                      bbox=dict(facecolor='white',edgecolor='white',pad= 0.,alpha=0.),zorder=7,fontdict = labelfont)
                #This should be the line to the label
                ax2.plot([time,last_label+xoffset],[vertical_start,vertical_start+offset],linest,color=color,linewidth=0.05,clip_on=False)
            prev_label = label



        #This is beyond stupid again, but hey it is python so needed to make things work.
        ax2.set_ylim([ax2miny,ax2maxy])
        fig.savefig(f"{Configuration['LOG_DIRECTORY']}ram_cpu.pdf")
        plt.close()

plot_usage_stats.__doc__ =f'''
 NAME:
    plot_usage_stats

 PURPOSE:
    Plot the RAM and CPU usage over time for the fit

 CATEGORY:
    write_functions

 INPUTS:
    Configuration = Standard FAT configuration

 OPTIONAL INPUTS:


 OUTPUTS:
    ram_cpu.pdf in the Logs directory

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def reorder_output_catalogue(Configuration,Full_Catalogue):
    #Read in the output catalogue
    with open(Configuration['OUTPUT_CATALOGUE']) as file:
        lines = file.readlines()
    #Determine column sizes
    header = lines[0]
    output = lines[1:]
    #IDs cannot have spaces
    outputIDs = []
    for line in output:
        outputIDs.append(line.split()[0].strip())
    #Sort based on the inpu IDs
    with open(Configuration['OUTPUT_CATALOGUE'],'w') as file:
        file.write(header)
        for galaxy in Full_Catalogue['DIRECTORYNAME']:
            print(galaxy)
            try:
                index_no = np.where(galaxy == \
                    np.array(outputIDs))[0][0]
                print(index_no)
                file.write(output[index_no])
            except IndexError:
                pass




reorder_output_catalogue.__doc__ =f'''
 NAME:
    reorder_output_catalogue

 PURPOSE:
    When running in multiprocessing mode the output catalogue can be
    in a different order as the input. This function makes sure the are sorted

 CATEGORY:
    write_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Full_Catalogue = Full input catalogue

 OPTIONAL INPUTS:


 OUTPUTS:
    reorder output catalogue

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def extract_date(string):
    tmp = string.split(' ')
    tmp2 = tmp[0].split('-')
    if len(tmp2) == 3:
        try:
            date =  datetime.strptime(f"{tmp[0]} {tmp[1]}", '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            date =  datetime.strptime(f"{tmp[0]} {tmp[1]}", '%Y-%m-%d %H:%M:%S')
    else:
        raise ProgramError("There is no date in the provided string.")
    return date

def sofia(template,name):
    with open(name,'w') as file:
        for key in template:
            if key[0] == 'E' or key [0] == 'H':
                file.write(template[key])
            else:
                file.write(f"{key} = {template[key]}\n")

sofia.__doc__ =f'''
 NAME:
    sofia
 PURPOSE:
    write a sofia2 dictionary into file
 CATEGORY:
    write_functions

 INPUTS:
    template = sofia template
    name = name of the file to write to

 OPTIONAL INPUTS:

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def tirific(Configuration,Tirific_Template, name = 'tirific.def', full_name = False  ):
    #IF we're writing we bump up the restart_ID and adjust the AZ1P angles to the current warping
    update_disk_angles(Configuration,Tirific_Template )
    try:
        Tirific_Template['RESTARTID'] = str(int(Tirific_Template['RESTARTID'])+1)
    except ValueError:
        Tirific_Template['RESTARTID'] = 0
    if full_name:
        file_name = name
    else:
        file_name = f'{Configuration["FITTING_DIR"]}{name}'
    with open(file_name, 'w') as file:
        for key in Tirific_Template:
            if key[0:5] == 'EMPTY':
                file.write('\n')
            else:
                file.write((f"{key}= {Tirific_Template[key]} \n"))
tirific.__doc__ =f'''
 NAME:
    tirific

 PURPOSE:
    Write a tirific template to file

 CATEGORY:
    write_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Tirific_Template = Standard FAT Tirific Template

 OPTIONAL INPUTS:


    name = 'tirific.def'
    name of the file to write to

 OUTPUTS:
    Tirific def file

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
 '''



# -*- coding: future_fstrings -*-
# This module contains a set of functions and classes that are used in several different Python scripts in the Database.
from astropy.io import fits
from astropy.wcs import WCS

from pyFAT_astro.Support.fat_errors import FunctionCallError,BadHeaderError,\
                                        BadCubeError,BadMaskError
import pyFAT_astro.Support.support_functions as sf

#from pyFAT_astro.Support.read_functions import obtain_border_pix
import pyFAT_astro
from scipy import ndimage
import numpy as np
import copy
import warnings
import os

#Check that the mask only contains the selected sources
def check_mask(Configuration,id,Fits_Files):
    sf.print_log(f'''CHECK_MASK: Checking the mask to contain only the correct source.
''',Configuration, case =['debug_start','verbose'])
    mask = fits.open(f"{Configuration['FITTING_DIR']}/{Fits_Files['MASK']}",uint = False, do_not_scale_image_data=True,ignore_blank = True, output_verify= 'ignore')

    if float(id) not in mask[0].data:
        sf.print_log(f'''CHECK_MASK: We cannot find the selected source in the mask. This will lead to errors. Aborting the fit.
''',Configuration,case=['main','screen'])
        BadMaskError(f" We can not find the sofia source id in the mask.")
    else:
        data = copy.deepcopy(mask[0].data)
        data[data != float(id)] = 0.
        diff = data-mask[0].data
        neg_index = np.where(diff < 0.)[0]
        if len(neg_index) != 0:
            sf.print_log(f'''CHECK_MASK: The initial mask had more than a single source. redoing the mask.
''',Configuration, case = ['verbose'])
            mask[0].data = data
            fits.writeto(f"{Configuration['FITTING_DIR']}{Fits_Files['MASK']}",mask[0].data,mask[0].header, overwrite = True)
# to ensure compatible units and calculations with th models we make the maps ourselves
    del mask[0].header['C*3']
    mask[0].data[mask[0].data> 0.5] = 1.
    chan_map = np.array(np.nansum(mask[0].data,axis=0),dtype =float)
    mask[0].header['BITPIX'] = -32
    fits.writeto(f"{Configuration['FITTING_DIR']}/Sofia_Output/{Configuration['BASE_NAME']}_chan.fits",chan_map,header=mask[0].header,overwrite = True)
    mask.close()
    if Configuration['SOFIA_RAN']:
        make_moments(Configuration, Fits_Files,fit_type='Generic_Initialize',vel_unit = 'm/s')
check_mask.__doc__ =f'''
 NAME:
    check_mask

 PURPOSE:
    Check that the mask only contains the source we want and not parts of other nearby sources.
    Subsequently make the channel_map and the moment maps.

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    id = the SoFiA ID of the source we are interested in
    Fits_Files = Standard FAT dictionary with filenames
    Tirific_Template = Standard FAT Tirific Template

 OPTIONAL INPUTS:
    fit_type = 'Undefined'

 OUTPUTS:
    moment maps and the channel map

 OPTIONAL OUTPUTS:
    a cleaned mask

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''




# Create a cube suitable for FAT
def create_fat_cube(Configuration, Fits_Files,sofia_catalogue=False,\
        id='No default',name='No default'):
    #First get the cubes
    if sofia_catalogue:
        Cube = fits.open(f"{Configuration['SOFIA_DIR']}{name}_{id}_cube.fits",\
            uint = False, do_not_scale_image_data=True,ignore_blank = True, output_verify= 'ignore')
    else:
        Cube = fits.open(Configuration['FITTING_DIR']+Fits_Files['ORIGINAL_CUBE'],\
            uint = False, do_not_scale_image_data=True,ignore_blank = True, output_verify= 'ignore')
    data = Cube[0].data
    hdr = Cube[0].header
    if hdr['NAXIS'] == 4:
        data = data[0,:,:,:]
        del hdr['NAXIS4']
        hdr['NAXIS'] = 3
    # clean the header
    hdr = sf.clean_header(Configuration,hdr)
    data = prep_cube(Configuration,hdr,data)
    # and write our new cube

    sf.print_log(f'''CREATE_FAT_CUBE: We are writing a FAT modified cube to be used for the fitting. This cube is called {Configuration['FITTING_DIR']+Fits_Files['FITTING_CUBE']}
''',Configuration)
    if sofia_catalogue:
        fits.writeto(f"{Configuration['MAIN_DIRECTORY']}{name}_FAT_cubelets/{name}_{id}/{name}_{id}_FAT.fits",data,hdr)
    else:
        fits.writeto(Configuration['FITTING_DIR']+Fits_Files['FITTING_CUBE'],data,hdr)
    # Release the arrays

    Cube.close()

    data = []
    hdr = []
create_fat_cube.__doc__ =f'''
 NAME:
    create_fat_cube

 PURPOSE:
    As we do not want to work from the original cube
    this function copies that cube into a new cube
    that we can then happily modify and work with.
    To keep disk space overhead this cube might later on also
    be cut down to size.

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames

 OPTIONAL INPUTS:

 OUTPUTS:
    The FAT input cube

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def cut_cubes(Configuration, Fits_Files, galaxy_box):
    sf.print_log(f'''CUT_CUBES: Starting to cut the cube_size.
''', Configuration, case= ['debug_start'])
    cube_edge= [6.,5.*round(Configuration['BEAM_IN_PIXELS'][0]),5.*round(Configuration['BEAM_IN_PIXELS'][0])]
    cube_size= []
    new_cube = []
    for i in [2,1,0]:
        cube_size.append(Configuration['NAXES'][i])
        new_cube.append([0,Configuration['NAXES'][i]])
    new_cube = np.array(new_cube,dtype=int)
    cut  = False
    for i,limit in enumerate(galaxy_box):
        if limit[0] > cube_edge[i]:
            cut = True
            new_cube[i,0] = limit[0]-int(cube_edge[i])
        if limit[1] < cube_size[i] - cube_edge[i]:
            cut = True
            new_cube[i,1] = limit[1]+int(cube_edge[i])
    if 'restart_fitting' in [x.lower() for x in Configuration['FITTING_STAGES']]:
        if 'create_fat_cube' in [x.lower() for x in Configuration['FITTING_STAGES']]:
            files_to_cut = [Fits_Files['FITTING_CUBE']]
        else:
            cut = False
    else:
        files_to_cut = [Fits_Files['FITTING_CUBE'],Fits_Files['MASK'],\
                        Fits_Files['MOMENT0'],\
                        Fits_Files['MOMENT1'],\
                        Fits_Files['MOMENT2'],\
                        Fits_Files['CHANNEL_MAP'],\
                        ]
    if cut:
        sf.print_log(f'''CUT_CUBES: Your input cube is significantly larger than the detected source.
{"":8s}CUT_CUBES: we will cut to x-axis = [{new_cube[2,0]},{new_cube[2,1]}] y-axis = [{new_cube[1,0]},{new_cube[1,1]}]
{"":8s}CUT_CUBES: z-axis = [{new_cube[2,0]},{new_cube[2,1]}].
{"":8s}CUT_CUBES: We will cut the following files:
{"":8s}CUT_CUBES: {', '.join(files_to_cut)}
''', Configuration)

        for file in files_to_cut:
            sf.print_log(f'''CUT_CUBES: We are cutting the file {file}
''', Configuration)
            if os.path.exists(f"{Configuration['FITTING_DIR']}{file}"):
                cutout_cube(Configuration,file,new_cube)
            else:
                if file == Fits_Files['CHANNEL_MAP']:
                    pass
                else:
                    raise FunctionCallError(f'We are trying to cut {file} but it does not exist')


    #We want to check if the cube has a decent number of pixels per beam.
    if not os.path.exists(f"{Configuration['FITTING_DIR']}/{Fits_Files['OPTIMIZED_CUBE']}"):
        optimized_cube(Configuration, Fits_Files)
    return new_cube
cut_cubes.__doc__ =f'''
 NAME:
    cut_cubes

 PURPOSE:
    Cut all FAT related output back in size to a system that fits snugly around the sofia detection

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames
    galaxy_box = array that contains the new size as
                 [[z_min,z_max],[y_min,y_max], [x_min,x_max]]
                 adhering to fits' idiotic way of reading fits files

 OPTIONAL INPUTS:

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    cutout_cube, optimized_cube

 NOTE: The cut cubes will overwrite the _FAT products
       down to size and overwrite the existing files,
       User supplied SoFIA products are hence copied as fat products.
       To keep them safe.
'''

def cutout_cube(Configuration,filename,sub_cube):
    Cube = fits.open(Configuration['FITTING_DIR']+filename,uint = False, do_not_scale_image_data=True,ignore_blank = True, output_verify= 'ignore')
    hdr = Cube[0].header

    if hdr['NAXIS'] == 3:
        data = Cube[0].data[sub_cube[0,0]:sub_cube[0,1],sub_cube[1,0]:sub_cube[1,1],sub_cube[2,0]:sub_cube[2,1]]
        hdr['NAXIS1'] = sub_cube[2,1]-sub_cube[2,0]
        hdr['NAXIS2'] = sub_cube[1,1]-sub_cube[1,0]
        hdr['NAXIS3'] = sub_cube[0,1]-sub_cube[0,0]
        hdr['CRPIX1'] = hdr['CRPIX1'] -sub_cube[2,0]
        hdr['CRPIX2'] = hdr['CRPIX2'] -sub_cube[1,0]
        hdr['CRPIX3'] = hdr['CRPIX3'] -sub_cube[0,0]
        #Only update when cutting the cube
        Configuration['NAXES'] =[ hdr['NAXIS1'],hdr['NAXIS2'],hdr['NAXIS3']]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            coordinate_frame = WCS(hdr)
            xlow,ylow,zlow = coordinate_frame.wcs_pix2world(1,1,1., 1.)
            xhigh,yhigh,zhigh = coordinate_frame.wcs_pix2world(*Configuration['NAXES'], 1.)
            xlim = np.sort([xlow,xhigh])
            ylim = np.sort([ylow,yhigh])
            zlim =np.sort([zlow,zhigh])/1000.
            sf.set_boundaries(Configuration,'VSYS',*zlim,input=True)
            sf.set_boundaries(Configuration,'XPOS',*xlim,input=True)
            sf.set_boundaries(Configuration,'YPOS',*ylim,input=True)

    elif hdr['NAXIS'] == 2:
        data = Cube[0].data[sub_cube[1,0]:sub_cube[1,1],sub_cube[2,0]:sub_cube[2,1]]
        hdr['NAXIS1'] = sub_cube[2,1]-sub_cube[2,0]
        hdr['NAXIS2'] = sub_cube[1,1]-sub_cube[1,0]
        hdr['CRPIX1'] = hdr['CRPIX1'] -sub_cube[2,0]
        hdr['CRPIX2'] = hdr['CRPIX2'] -sub_cube[1,0]

    Cube.close()
    fits.writeto(Configuration['FITTING_DIR']+filename,data,hdr,overwrite = True)
cutout_cube.__doc__ =f'''
 NAME:
    cutout_cube

 PURPOSE:
    Cut filename back to the size of subcube, update the header and write back to disk.

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    filename = name of the cube to be cut
    sub_cube = array that contains the new size as
                [[z_min,z_max],[y_min,y_max], [x_min,x_max]]
                adhering to fits' idiotic way of reading fits files.

 OPTIONAL INPUTS:

 OUTPUTS:
    the cut cube is written to disk.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

# Extract a PV-Diagrams
def extract_pv(Configuration,cube_in,angle,center=[-1,-1,-1],finalsize=[-1,-1],convert=-1):
    sf.print_log(f'''EXTRACT_PV: We are the extraction of a PV-Diagram
{'':8s} PA = {angle}
{'':8s} center = {center}
{'':8s} finalsize = {finalsize}
{'':8s} convert = {convert}
''', Configuration, case = ['debug_start'])


    cube = copy.deepcopy(cube_in)
    hdr = copy.deepcopy(cube[0].header)
    TwoD_hdr= copy.deepcopy(cube[0].header)
    data = copy.deepcopy(cube[0].data)
    #Because astro py is even dumber than Python
    try:
        if hdr['CUNIT3'].lower() == 'km/s':
            hdr['CUNIT3'] = 'm/s'
            hdr['CDELT3'] = hdr['CDELT3']*1000.
            hdr['CRVAL3'] = hdr['CRVAL3']*1000.
        elif hdr['CUNIT3'].lower() == 'm/s':
            hdr['CUNIT3'] = 'm/s'
    except KeyError:
        hdr['CUNIT3'] = 'm/s'
    if center[0] == -1:
        center = [hdr['CRVAL1'],hdr['CRVAL2'],hdr['CRVAL3']]
        xcenter,ycenter,zcenter = hdr['CRPIX1'],hdr['CRPIX2'],hdr['CRPIX3']
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            coordinate_frame = WCS(hdr)
        xcenter,ycenter,zcenter = coordinate_frame.wcs_world2pix(center[0], center[1], center[2], 1.)

    nz, ny, nx = data.shape
    if finalsize[0] != -1:
        if finalsize[1] >nz:
            finalsize[1] = nz
        if finalsize[0] > nx:
            finalsize[0] = nx
    # if the center is not set assume the crval values

    sf.print_log(f'''EXTRACT_PV: The shape of the output
{'':8s} nz = {nz}
{'':8s} ny = {ny}
{'':8s} nx = {nx}
''', Configuration,case=['verbose'])
    x1,x2,y1,y2 = sf.obtain_border_pix(Configuration,angle,[xcenter,ycenter])
    linex,liney,linez = np.linspace(x1,x2,nx), np.linspace(y1,y2,nx), np.linspace(0,nz-1,nz)
    #This only works when ny == nx hence nx is used in liney
    new_coordinates = np.array([(z,y,x)
                        for z in linez
                        for y,x in zip(liney,linex)
                        ],dtype=float).transpose().reshape((-1,nz,nx))
    #spatial_resolution = abs((abs(x2-x1)/nx)*np.sin(np.radians(angle)))+abs(abs(y2-y1)/ny*np.cos(np.radians(angle)))
    PV = ndimage.map_coordinates(data, new_coordinates,order=1)
    if hdr['CDELT1'] < 0:
        PV = PV[:,::-1]

    if finalsize[0] == -1:
        # then lets update the header
        # As python is stupid making a simple copy will mean that these changes are still applied to hudulist
        TwoD_hdr['NAXIS2'] = nz
        TwoD_hdr['NAXIS1'] = nx

        TwoD_hdr['CRPIX2'] = hdr['CRPIX3']
        TwoD_hdr['CRPIX1'] = xcenter+1
    else:
        zstart = sf.set_limits(int(zcenter-finalsize[1]/2.),0,int(nz))
        zend = sf.set_limits(int(zcenter+finalsize[1]/2.),0,int(nz))
        xstart = sf.set_limits(int(xcenter-finalsize[0]/2.),0,int(nx))
        xend = sf.set_limits(int(xcenter+finalsize[0]/2.),0,int(nx))

        PV =  PV[zstart:zend, xstart:xend]
        TwoD_hdr['NAXIS2'] = int(finalsize[1])
        TwoD_hdr['NAXIS1'] = int(finalsize[0])
        TwoD_hdr['CRPIX2'] = hdr['CRPIX3']-int(nz/2.-finalsize[1]/2.)
        TwoD_hdr['CRPIX1'] = int(finalsize[0]/2.)+1

    if convert !=-1:
        TwoD_hdr['CRVAL2'] = hdr['CRVAL3']/convert
        TwoD_hdr['CDELT2'] = hdr['CDELT3']/convert
    else:
        TwoD_hdr['CRVAL2'] = hdr['CRVAL3']
        TwoD_hdr['CDELT2'] = hdr['CDELT3']
    TwoD_hdr['CTYPE2'] = hdr['CTYPE3']
    try:
        if hdr['CUNIT3'].lower() == 'm/s' and convert == 1000. :
            TwoD_hdr['CDELT2'] = hdr['CDELT3']/1000.
            TwoD_hdr['CRVAL2'] = hdr['CRVAL3']/1000.
            TwoD_hdr['CUNIT2'] = 'km/s'
            del (TwoD_hdr['CUNIT3'])
        elif  convert != -1:
            del (TwoD_hdr['CUNIT3'])
            try:
                del (TwoD_hdr['CUNIT2'])
            except KeyError:
                pass
        else:
            TwoD_hdr['CUNIT2'] = hdr['CUNIT3']
            del (TwoD_hdr['CUNIT3'])
    except KeyError:
        sf.print_log(f'''EXTRACT_PV: We could not find units in the header for the 3rd axis.
''', Configuration)
    del (TwoD_hdr['CRPIX3'])
    del (TwoD_hdr['CRVAL3'])
    del (TwoD_hdr['CDELT3'])
    del (TwoD_hdr['CTYPE3'])
    del (TwoD_hdr['NAXIS3'])
    TwoD_hdr['NAXIS'] = 2
    TwoD_hdr['CRVAL1'] = 0.
    #Because we used nx in the linspace for liney we also use it here
    TwoD_hdr['CDELT1'] = np.sqrt(((x2-x1)*abs(hdr['CDELT1'])/nx)**2+((y2-y1)*abs(hdr['CDELT2'])/nx)**2)*3600.

    TwoD_hdr['CTYPE1'] = 'OFFSET'
    TwoD_hdr['CUNIT1'] = 'ARCSEC'
    TwoD_hdr['HISTORY'] = f'EXTRACT_PV: PV diagram extracted with angle {angle} and center {center}'
    # Then we change the cube and rteturn the PV construct
    cube[0].header = TwoD_hdr
    cube[0].data = PV

    return cube
extract_pv.__doc__ = '''
 NAME:
    extract_pv

 PURPOSE:
    extract a PV diagram from a cube object. Angle is the PA and center the central location. The profile is extracted over the full length of the cube and afterwards cut back to the finalsize.

 CATEGORY:
     fits_functions

 INPUTS:
    Configuration = Standard FAT Configuration
    cube_in = is a fits cube object
    angle = Pa of the slice

 OPTIONAL INPUTS:
    center = [-1,-1,-1]
    the central location of the slice in WCS map_coordinates [RA,DEC,VSYS], default is the CRVAL values in the header

    finalsize = [-1,-1,-1]
    final size of the PV-diagram in pixels, default is no cutting

    convert=-1
    conversion factor for velocity axis, default no conversion

 KEYWORD PARAMETERS:

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''


def prep_cube(Configuration,hdr,data):
    sf.print_log( f'''PREPROCESSING: starting the preprocessing of the cube
''',Configuration, case = ['debug_start'])

    if hdr['CDELT3'] < -1:
        sf.print_log( f'''PREPROCESSING: Your velocity axis is declining with increasing channels
{"":8s}PREPROCESSING: We reversed the velocity axis.
''', Configuration)
        hdr['CDELT3'] = abs(hdr['CDELT3'])
        hdr['CRPIX3'] = hdr['NAXIS3']-hdr['CRPIX3']+1
        data = data[::-1,:,:]
    #Check for zeros
    if np.where(data == 0.):
        sf.print_log(f'''PREPROCESSING: Your cube contains values exactly 0. If this is padding these should be blanks.
{"":8s}PREPROCESSING: We have changed them to blanks.
''', Configuration)
        data[np.where(data == 0.)] = float('NaN')

    # check for blank channels and noise statistics
    cube_ok = False
    prev_first_comparison = 0.
    prev_last_comparison = 0.
    times_cut_first_channel = 0
    times_cut_last_channel = 0
    corner_box_size = int(np.floor(2*hdr['BMAJ']/abs(hdr['CDELT1'])))
    if corner_box_size > np.mean([hdr['NAXIS1'],hdr['NAXIS2']])/10.:
        corner_box_size=int(np.floor(np.mean([hdr['NAXIS1'],hdr['NAXIS2']])/10.))
    while not cube_ok:
        while np.isnan(data[0,:,:]).all():
            data=data[1:,:,:]
            hdr['NAXIS3'] = hdr['NAXIS3']-1
            hdr['CRPIX3'] = hdr['CRPIX3'] - 1
            if hdr['NAXIS3'] < 5:
                sf.print_log(f'''PREPROCESSING: This cube has too many blanked channels.
''', Configuration,case = ['main', 'screen'])
                raise BadCubeError("The cube has too many blanked channels")
            log_statement = f'''PREPROCESSING: We are cutting the cube as the first channel is completely blank.
'''
            sf.print_log(f'''PREPROCESSING: We are cutting the cube as the first channel is completely blank.
''', Configuration)
        while np.isnan(data[-1,:,:]).all():
            data=data[:-1,:,:]
            hdr['NAXIS3'] = hdr['NAXIS3']-1
            if hdr['NAXIS3'] < 5:
                sf.print_log(f'''PREPROCESSING: This cube has too many blanked channels.
''', Configuration,case = ['main', 'screen'])
                raise BadCubeError("The cube has too many blanked channels")
            sf.print_log(f'''PREPROCESSING: We are cutting the cube as the last channel is completely blank.
''', Configuration)
        #Then check the noise statistics
        noise_first_channel = np.nanstd(data[0,:,:])
        noise_last_channel = np.nanstd(data[-1,:,:])
        noise_bottom_right =  np.nanstd(data[:,:corner_box_size,-corner_box_size:])
        noise_top_right =  np.nanstd(data[:,-corner_box_size:,-corner_box_size:])
        noise_bottom_left =  np.nanstd(data[:,:corner_box_size,:corner_box_size])
        noise_top_left =  np.nanstd(data[:,-corner_box_size:,:corner_box_size])
        noise_corner = np.nanmean([noise_top_right,noise_bottom_right,noise_bottom_left,noise_top_left])
        channel_noise = np.nanmean([noise_first_channel,noise_last_channel])
        if ~np.isfinite(noise_corner):
            noise_corner = copy.deepcopy(channel_noise)
        difference = abs((noise_first_channel-noise_last_channel)/noise_first_channel)
        if noise_corner/channel_noise >1.5 and difference < 0.2:
            noise_corner = copy.deepcopy(channel_noise)
        difference2 = abs(channel_noise-noise_corner)/channel_noise
        if difference < 0.2 and np.isfinite(difference) and difference2 < 0.25:
            cube_ok = True
        else:
            sf.print_log(f'''PREPROCESSING: We are cutting the cube as clearly the noise statistics are off.
{"":8s}PREPROCESSING: Noise in the first channel is {noise_first_channel}
{"":8s}PREPROCESSING: Noise in the last channel is {noise_last_channel}
{"":8s}PREPROCESSING: Noise in the corners is {noise_corner}
''',Configuration,case= ['verbose'])
            first_comparison = abs(noise_first_channel-noise_corner)/noise_corner
            last_comparison = abs(noise_last_channel-noise_corner)/noise_corner
            if prev_first_comparison == 0:
                prev_first_comparison = first_comparison
            if prev_last_comparison == 0.:
                prev_last_comparison = last_comparison
            if (first_comparison > last_comparison and \
               abs((first_comparison-prev_first_comparison)/first_comparison) >= abs((last_comparison-prev_last_comparison)/last_comparison) and \
               times_cut_first_channel < 0.1*hdr['NAXIS3']) or \
               ~np.isfinite(noise_first_channel):
                prev_first_comparison = first_comparison
                times_cut_first_channel += 1
                data=data[1:,:,:]
                hdr['CRPIX3'] = hdr['CRPIX3'] - 1
                hdr['NAXIS3'] = hdr['NAXIS3'] - 1
            else:
                if times_cut_last_channel < 0.1*hdr['CRPIX3'] or \
                    ~np.isfinite(noise_last_channel):
                    prev_last_comparison = last_comparison
                    times_cut_last_channel += 1
                    data=data[:-1,:,:]
                    hdr['NAXIS3'] = hdr['NAXIS3'] - 1
                else:
                    cube_ok = True

            if times_cut_first_channel >= 8 and times_cut_last_channel >= 8:
                cube_ok = True
                sf.print_log( f'''PREPROCESSING: This cube has non-uniform noise statistics.'
''',Configuration)
            if hdr['NAXIS3'] < 5:
                sf.print_log(f'''PREPROCESSING: This cube has noise statistics that cannot be dealt with.
''',Configuration,case = ['main','screen'])
                raise BadCubeError('The Cube has noise statistics that cannot be dealt with')
    if ~np.isfinite(noise_corner):
        sf.print_log(f'''PREPROCESSING: This cube has noise statistics that cannot be dealt with.
''',Configuration,case = ['main','screen'])
        raise BadCubeError('The Cube has noise statistics that cannot be dealt with')
    #hdr['FATNOISE'] = noise_corner
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        low_noise_indices = np.where(data < -10*noise_corner)
    if len(low_noise_indices) > 0.0001*hdr['NAXIS1']*hdr['NAXIS2']*hdr['NAXIS3']:
        data[low_noise_indices] = float('NaN')
        sf.print_log(f'''PREPROCESSING: Your cube had a significant amount of values below -10*sigma. If you do not have a central absorption source there is something seriously wrong with the cube.
{"":8s}PREPROCESSING: We blanked these values.
''',Configuration)

    #Let's make sure that a central absorption source is treated the same regardless od the size of the cube
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        buffer=5.*hdr['BMAJ']/np.mean([abs(hdr['CDELT1']),abs(hdr['CDELT2'])])
        central = data[:,int(hdr['NAXIS2']/2.-buffer):int(hdr['NAXIS2']/2.+buffer),\
                        int(hdr['NAXIS1']/2.-buffer):int(hdr['NAXIS1']/2.+buffer)]

        central_noise_indices = np.where(central < -10*noise_corner)
    if len(central_noise_indices) > 0.0001*4*buffer*hdr['NAXIS3']:
        central[central_noise_indices] = float('NaN')
        data[:,int(hdr['NAXIS2']/2.-buffer):int(hdr['NAXIS2']/2.+buffer),\
                        int(hdr['NAXIS1']/2.-buffer):int(hdr['NAXIS1']/2.+buffer)] =\
                        central
        sf.print_log(f'''PREPROCESSING: Your cube had a significant amount of values below -10*sigma. If you do not have a central absorption source there is something seriously wrong with the cube.
{"":8s}PREPROCESSING: We blanked these values.
''',Configuration)


    # Check whether any channels got fully blanked
    blanked_channels = []
    for z in range(hdr['NAXIS3']):
        if np.isnan(data[z,:,:]).all():
            blanked_channels.append(str(z))
    if len(blanked_channels) == 0.:
        blanked_channels = ['-1']
    hdr['BL_CHAN'] = ','.join(blanked_channels)
    #Previously SoFiA was not dealing with blanks properly and the statement went here
    #finally we want to get the noise level from the negative in the final cube

    new_noise = np.nanstd(np.hstack([data[data<0],-1.*data[data<0]]))
    #If we have noiseles cubes this will be far to low
    if abs(new_noise/np.mean([channel_noise,noise_corner])) > 4. \
        or abs(new_noise/np.mean([channel_noise,noise_corner])) < 0.25:
            sf.print_log(f'''PREPROCESSING: There is something odd in the noise statistics of your cube. We are not using the nehgative values
''',Configuration)
            #new_noise = np.mean([channel_noise,noise_corner])
    hdr['FATNOISE'] = new_noise

    return data
prep_cube.__doc__ =f'''
 NAME:
    prep_cube

 PURPOSE:
    Clean up the cube  and make sure it has all the right
    characteristics and blanks

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    hdr = header to be cleaned
    data = data array to be cleaned

 OPTIONAL INPUTS:

 OUTPUTS:
    the updated header

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

#Create an optimized cube if required
def optimized_cube(Configuration,Fits_Files):
    pix_per_beam = Configuration['BEAM_IN_PIXELS'][1]
    if pix_per_beam > Configuration['OPT_PIXEL_BEAM']:
        cube = fits.open(Configuration['FITTING_DIR']+Fits_Files['FITTING_CUBE'])
        data = cube[0].data
        hdr = cube[0].header
        if f"{abs(hdr['CDELT1']):.16f}" != f"{abs(hdr['CDELT2']):.16f}":
            sf.print_log(f'''OPTIMIZED_CUBE: Your input cube does not have square pixels.
{"":8s}OPTIMIZED_CUBE: FAT cannot optimize your cube.
''', Configuration)
        cube.close()
        required_cdelt = hdr['BMIN']/int(Configuration['OPT_PIXEL_BEAM'])
        ratio = required_cdelt/abs(hdr['CDELT2'])
        opt_data,opt_hdr = regrid_cube(data, hdr, ratio)

        fits.writeto(Configuration['FITTING_DIR']+Fits_Files['OPTIMIZED_CUBE'], opt_data,opt_hdr)

        Configuration['OPTIMIZED'] = True
        sf.print_log(f'''OPTIMIZED_CUBE: Your input cube has {pix_per_beam} pixels along the minor FWHM.
{"":8s}OPTIMIZED_CUBE: You requested { Configuration['OPT_PIXEL_BEAM']} therefore we regridded the cube into a new cube.
{"":8s}OPTIMIZED_CUBE: We are using the pixel size of {required_cdelt}.
''', Configuration)

    else:
        sf.print_log(f'''OPTIMIZED_CUBE: Your input cube has {pix_per_beam} pixels along the minor FWHM.
{"":8s}OPTIMIZED_CUBE: You requested {Configuration['OPT_PIXEL_BEAM']} but we cannot improve the resolution.
{"":8s}OPTIMIZED_CUBE: We are using the pixel size of the original cube.
''', Configuration)
optimized_cube.__doc__ =f'''
 NAME:
    optimized_cube

 PURPOSE:
    Check the amount of pixels that are in the minor axis FWHM,
    if more than OPT_PIXELBEAM regrid the cube to less pixels
    and write FAT_opt cube and set Configuration['OPTIMIZED'] = True

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames

 OPTIONAL INPUTS:

 OUTPUTS:
    An optimized cube is created and written to disk.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def make_moments(Configuration,Fits_Files,fit_type = 'Undefined',
                 moments = [0,1,2],overwrite = False, level=None,
                  vel_unit= None):
    sf.print_log(f'''MAKE_MOMENTS: We are starting to create the moment maps.
''',Configuration, case = ['debug_start'])
    if fit_type == 'Generic_Initialize':
        filename = f"{Configuration['FITTING_DIR']}{Fits_Files['FITTING_CUBE']}"
        basename = f"{Configuration['BASE_NAME']}"
        directory = f"{Configuration['FITTING_DIR']}/Sofia_Output/"
        mask_cube = f"{Configuration['FITTING_DIR']}/{Fits_Files['MASK']}"
    elif fit_type == 'Generic_Final':
        filename = f"{Configuration['FITTING_DIR']}/Finalmodel/Finalmodel.fits"
        basename = 'Finalmodel'
        directory = f"{Configuration['FITTING_DIR']}/Finalmodel/"
        mask_cube = f"{Configuration['FITTING_DIR']}/{Fits_Files['MASK']}"
    else:
        filename = f"{Configuration['FITTING_DIR']}{fit_type}/{fit_type}.fits"
        basename = fit_type
        directory = f"{Configuration['FITTING_DIR']}{fit_type}"
        if not level:
            mask_cube = f"{Configuration['FITTING_DIR']}/{Fits_Files['MASK']}"

    cube = fits.open(filename)
    if vel_unit:
        cube[0].header['CUNIT3'] = vel_unit
    if mask_cube:
        mask = fits.open(mask_cube)
        with np.errstate(invalid='ignore', divide='ignore'):
            cube[0].data[mask[0].data < 0.5] = float('NaN')
        mask.close()
    else:
        if not level:
            level = 3.*np.mean([np.nanstd(cube[0].data[0:2,:,:]),np.nanstd(cube[0].data[-3:-1,:,:])])
        with np.errstate(invalid='ignore', divide='ignore'):
            cube[0].data[cube[0].data < level] = float('NaN')
    try:
        if cube[0].header['CUNIT3'].lower().strip() == 'm/s':
            sf.print_log(f"MAKE_MOMENTS: We convert your m/s to km/s. \n", Configuration, case=['verbose'])
            cube[0].header['CUNIT3'] = 'km/s'
            cube[0].header['CDELT3'] = cube[0].header['CDELT3']/1000.
            cube[0].header['CRVAL3'] = cube[0].header['CRVAL3']/1000.
        elif cube[0].header['CUNIT3'].lower().strip() == 'km/s':
            pass
        else:
            sf.print_log(f"MAKE_MOMENTS: Your Velocity unit {cube[0].header['CUNIT3']} is weird. Your units could be off", Configuration)
    except KeyError:
        sf.print_log(f"MAKE_MOMENTS: Your CUNIT3 is missing, that is bad practice. We'll add a blank one but we're not guessing the value", Configuration)
        cube[0].header['CUNIT3'] = 'Unknown'
    #Make a 2D header to use
    hdr2D = copy.deepcopy(cube[0].header)
    hdr2D.remove('NAXIS3')
    hdr2D['NAXIS'] = 2
    # removing the third axis means we cannot correct for varying platescale, Sofia does so this is and issue so let's not do this
    hdr2D.remove('CDELT3')
    hdr2D.remove('CTYPE3')
    hdr2D.remove('CUNIT3')
    hdr2D.remove('CRPIX3')
    hdr2D.remove('CRVAL3')

    # we need a moment 0 for the moment 2 as well
    if 0 in moments:
        hdr2D['BUNIT'] = f"{cube[0].header['BUNIT']}*{cube[0].header['CUNIT3']}"
        moment0 = np.nansum(cube[0].data, axis=0) * cube[0].header['CDELT3']
        moment0[np.invert(np.isfinite(moment0))] = float('NaN')
        try:
            hdr2D['DATAMAX'] = np.nanmax(moment0)
            hdr2D['DATAMIN'] = np.nanmin(moment0)
            fits.writeto(f"{directory}/{basename}_mom0.fits",moment0,hdr2D,overwrite = overwrite)
        except ValueError:
            sf.print_log(f"MAKE_MOMENTS: Your Moment 0 has bad data and we could not write the moment 0 fits file", Configuration)


        
    if 1 in moments or 2 in moments:
        zaxis = cube[0].header['CRVAL3'] + (np.arange(cube[0].header['NAXIS3'])+1 \
              - cube[0].header['CRPIX3']) * cube[0].header['CDELT3']
        c=np.transpose(np.resize(zaxis,[cube[0].header['NAXIS1'],cube[0].header['NAXIS2'],len(zaxis)]),(2,1,0))
        hdr2D['BUNIT'] = f"{cube[0].header['CUNIT3']}"
        # Remember Python is stupid so z,y,x
        with np.errstate(invalid='ignore', divide='ignore'):
            moment1 = np.nansum(cube[0].data*c, axis=0)/ np.nansum(cube[0].data, axis=0)
        moment1[np.invert(np.isfinite(moment1))] = float('NaN')
        try:
            hdr2D['DATAMAX'] = np.nanmax(moment1)
            hdr2D['DATAMIN'] = np.nanmin(moment1)
            if 1 in moments:
                fits.writeto(f"{directory}/{basename}_mom1.fits",moment1,hdr2D,\
                    overwrite = overwrite)

        except ValueError:
            sf.print_log(f"MAKE_MOMENTS: Your Moment 1 has bad data and we could not write the moment 0 fits file", Configuration)

        if 2 in moments:
            d = c - np.resize(moment1,[len(zaxis),cube[0].header['NAXIS2'],cube[0].header['NAXIS1']])
            with np.errstate(invalid='ignore', divide='ignore'):
                moment2 = np.sqrt(np.nansum(cube[0].data*d**2, axis=0)/ np.nansum(cube[0].data, axis=0))
            moment2[np.invert(np.isfinite(moment1))] = float('NaN')
            try:
                hdr2D['DATAMAX'] = np.nanmax(moment2)
                hdr2D['DATAMIN'] = np.nanmin(moment2)
                fits.writeto(f"{directory}/{basename}_mom2.fits",\
                    moment2,hdr2D,overwrite = overwrite)

            except ValueError:
                sf.print_log(f"MAKE_MOMENTS: Your Moment 1 has bad data and we could not write the moment 0 fits file", Configuration)

    cube.close()

make_moments.__doc__ =f'''
 NAME:
    make_moments

 PURPOSE:
    Make the moment maps

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames


 OPTIONAL INPUTS:
    fit_type = 'Undefined'
    type of ftting

    moments = [0,1,2]
    moment maps to create

    overwrite = False
    overwrite existing maps

    level=None
    cutoff level to use, if set the mask will not be used

    vel_unit= none
    velocity unit of the input cube

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def regrid_cube(data,hdr,ratio):
    regrid_hdr = copy.deepcopy(hdr)
    # First get the shape of the data
    shape = np.array(data.shape, dtype=float)
    #The new shape has to be integers
    new_shape = [int(x/ratio) for x in shape]
    # Which means our real ratio is
    real_ratio = shape/new_shape
    #* np.ceil(shape / ratio).astype(int)
    new_shape[0] = shape[0]
    # Create the zero-padded array and assign it with the old density
    regrid_data = regridder(data,new_shape)
    regrid_hdr['CRPIX1'] =   (regrid_hdr['CRPIX1']-1)/real_ratio[2]+1
    regrid_hdr['CRPIX2'] =   (regrid_hdr['CRPIX2']-1)/real_ratio[1]+1
    wcs_found = False
    try:
        regrid_hdr['CDELT1'] =   regrid_hdr['CDELT1']*real_ratio[2]
        regrid_hdr['CDELT2'] =   regrid_hdr['CDELT2']*real_ratio[1]
        wcs_found = True
    except KeyError:
        print("No CDELT found")
    try:
        regrid_hdr['CD1_1'] =   regrid_hdr['CD1_1']*real_ratio[2]
        regrid_hdr['CD2_2'] =   regrid_hdr['CD2_2']*real_ratio[1]
        wcs_found = True
    except KeyError:
        if not wcs_found:
            print("No CD corr matrix found")
    try:
        regrid_hdr['CD1_2'] =   regrid_hdr['CD1_2']*real_ratio[2]
        regrid_hdr['CD2_1'] =   regrid_hdr['CD2_1']*real_ratio[1]
        wcs_found = True
    except KeyError:
        if not wcs_found:
            print("No CD cross-corr matrix found")

    regrid_hdr['NAXIS1'] =   new_shape[2]
    regrid_hdr['NAXIS2'] =   new_shape[1]
    return regrid_data,regrid_hdr
regrid_cube.__doc__ =f'''
 NAME:
    regrid_cube

 PURPOSE:
    Regrid a cube to an arbitrary smaller cube in the spatial plane with larger pixels

 CATEGORY:
    fits_functions

 INPUTS:
    data = the array containing the input cube.
    hdr = the header data
    ratio = the requested ratio of old/new

 OPTIONAL INPUTS:

 OUTPUTS:
    regriddata = regridded data array
    regrid_hdr = header corresponding to the regridded data

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    copy.deepcopy, regridder, np.array

 NOTE: This might not work well with extended projections.
'''

def regridder(oldarray, newshape):
    oldshape = np.array(oldarray.shape)
    newshape = np.array(newshape, dtype=float)
    ratios = oldshape/newshape
        # calculate new dims
    nslices = [ slice(0,j) for j in list(newshape) ]
    #make a list with new coord
    new_coordinates = np.mgrid[nslices]
    #scale the new coordinates
    for i in range(len(ratios)):
        new_coordinates[i] *= ratios[i]
    #create our regridded array
    newarray = ndimage.map_coordinates(oldarray, new_coordinates,order=1)

    return newarray
regridder.__doc__ =f'''
 NAME:
regridder
 PURPOSE:
Regrid an array into a new shape through the ndimage module
 CATEGORY:
    fits_functions

 INPUTS:
    oldarray = the larger array
    newshape = the new shape that is requested

 OPTIONAL INPUTS:

 OUTPUTS:
    newarray = regridded array

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    scipy.ndimage.map_coordinates, np.array, np.mgrid

 NOTE:
'''
